"""
Manifest
========

Types:
------
  + Torch types:
    - TorchTensor
    - TorchModule  (validation only)
    - TorchGenerator

JSON encoders
-------------
  + Torch types:
    - torch.Tensor
    - torch.nn.module (provided as a pair of methods: torch_module_state_encoder, torch_module_state_decoder)
    - torch.Generator
"""

from __future__ import annotations

from typing import Tuple, Dict

from .base import Serializable, ABCSerializable, json_like
from .base_types import SerializedData
from .numpy import Array

import numpy  # There are sometimes issues when torch is imported before numpy
import torch
import torch.nn as nn

# TODO: Allow indexing as with Array, to specify dtype & ndim
class TorchTensor(Serializable, torch.Tensor):
    class Data(SerializedData):
        data: Array
        def encode(tensor): return tensor.cpu().detach().numpy()
        def decode(data): return torch.tensor(data.data)

Tensor = TorchTensor

type_to_register = torch.Tensor
cls = TorchTensor
ABCSerializable.register(type_to_register)
ABCSerializable._base_types[cls] = type_to_register
for C in cls.mro():
    if issubclass(C, Serializable):
        C._registry[type_to_register] = cls

class TorchModule(nn.Module):
    """
    Validation only type; for serializing a PyTorch model, see `scityping.ml.TorchModel`.
    """
    @classmethod
    def __get_validators__(cls):
        yield cls.validate
    @classmethod
    def validate(cls, v):
        if not isinstance(v, nn.Module):
            raise TypeError("Expected a torch.nn.Module; received "
                            f"'{v}' (type: {type(v)})")
        return v

Module = TorchModule

TorchModuleState = Tuple[str, Dict[str, Array]]

# TODO: Package into a TorchModuleState class
def torch_module_state_decoder(v: Tuple[str, Dict[str,str]]):
    """
    Decode a torch module state serialized with `torch_module_state_reduce`.
    """
    if not json_like(v, "TorchModuleState"):
        raise TypeError("Argument is not a serialized torch state. "
                        f"Received: {v}.")
    encoded_state = v[1]
    # State is encoded as serialized arrays, so for each we have to:
    # 1. Deserialize the array  2. Convert to a pytorch tensor
    state = {param: torch.tensor(array) for param, array in encoded_state.items()}
    return state

def torch_module_state_encoder(
      v: nn.Module, compression: str='blosc', encoding: str='b85'
    ) -> Tuple[str, Dict[str,str]]:
    """
    Encode to PyTorch model state into an ASCII string, which can
    be included in a JSON file.

    .. Note:: This only encodes the model _state_ (i.e. the value
       from `v.state_dict()`). To reconstruct the model, one also
       needs to store the code, or some other means to reconstruct
       the architecture.

    Parameters
    ----------
    v: PyTorch model
    compression: 'blosc' | 'none'
       Which routine to use to compress bytes, if any.
    encoding: 'b85'
       Which ASCII encoding to use. Currently only 'b85' is supported.
       The `base64` module is used for this encoding.

    Returns
    -------
    ``("TorchModuleState", <data>)``
    where `<data>` is a dictionary with four fields:
       - encoding: Value of `encoding`
       - compression: Value of `compression`
       - description: Free-form, human-readable summary of the model
           (obtained by ``str(v)``)
       - state: The encoded model state.

    """
    if not isinstance(v, nn.Module):
        raise TypeError("This JSON encoder is only intended for PyTorch "
                        f"modules (received value of type {type(v)}).")
    state = {param: tensor.cpu().detach().numpy()
             for param, tensor in v.state_dict().items()}

    return ("TorchModuleState", state)


class TorchGenerator(Serializable, torch.Generator):
    """
    Pydantic-aware torch Generator.
    Since random number generators are platform dependent, serializing-
    deserializing from one machine to another, or from CPU to GPU, is likely
    to produce different results. However on the same machine and device,
    results should be consistent.
    """
    class Data(SerializedData):
        device: str
        state: TorchTensor
        def encode(gen): return gen.device.type, gen.get_state()
        def decode(data):
            try:
                gen = torch.Generator(data.device)
            except RuntimeError:
                gen = torch.Generator()
                logger.error(
                    "Unable to instantiate a torch Generator on device "
                    f"'{data.device}'. Used the default '{gen.device}' instead.")
            # NB: It seems to be a bug in PyTorch that even for a torch generator
            #     located on GPU, `set_state` only accepts arguments located on CPU.
            #     Otherwise, presumably the error message
            #     – "RNG state must be a torch.ByteTensor" – would be more useful.
            gen.set_state(data.state.to("cpu"))
            return gen

Generator = TorchGenerator

type_to_register = torch.Generator
cls = TorchGenerator
ABCSerializable.register(type_to_register)
ABCSerializable._base_types[cls] = type_to_register
for C in cls.mro():
    if issubclass(C, Serializable):
        C._registry[type_to_register] = cls



from __future__ import annotations

import abc
import logging
import textwrap
import inspect
from typing import get_type_hints, ClassVar, Union, Type, Any, Literal, List, Tuple
from collections import ChainMap
from collections.abc import (
    Callable as Callable_, Sequence as Sequence_, Iterable as Iterable_,
    Mapping as Mapping_)
from dataclasses import fields, is_dataclass, MISSING
from types import FunctionType
from .utils import get_type_key, TypeRegistry
from .typing import StrictStr, StrictBytes, StrictInt, StrictFloat, StrictBool

try:
    # v3.10+ allows notation like ``int|str``, which instantiates a `UnionType`.
    # We treat these as equivalent to `Union[int,str]`
    from types import UnionType
except:
    class UnionType:  # Dummy type, so that `instance(T, UnionType)` doesn’t raise an exception but returns False
        pass

try:
    from pydantic import BaseModel
except ModuleNotFoundError:
    # BaseModel is used only for isinstance checks – if pydantic is not loaded,
    # those tests are False by default. Therefore instantiating a dummy class suffices
    class BaseModel:
        def __new__(cls):
            raise RuntimeError("This dummy BaseModel is not meant to be instantiated.")

logger = logging.getLogger(__name__)

class MISSING:  # Sentinel value when looking for a value in the namespace
    pass

__all__ = ["json_like", "Serializable", "Serialized", "Dataclass"]

# ##############
# Custom JSON objects
# We use the following convention for JSON serializations: objects are
# serialized as tuples, where the first entry is a unique string identifying
# type. Thus we can check this string to know whether we should attempt decoding.

# TODO: Rename to `reduce_like` (or something along those lines)
def json_like(value: Any, type_str: Union[str,List[str]],
              case_sensitive: bool=False):
    """
    Convenience fonction for checking whether a serialized value might be a
    object serialized with `scityping.base.reduce`.

    Args:
        value: The value for which we want to determine if it is a
            JSON-serialized object.
        type_str: The type string of the type we are attempting to
            deserialize into. May be in list, in which the function returns true
            if any one matches.
        case_sensitive: Whether the comparison to `type_str` should be
            case-sensitive.
    """
    if isinstance(type_str, str):
        type_str = [type_str]
    casefold = (lambda v: v) if case_sensitive else str.casefold
    return any(
        (not isinstance(value, str) and isinstance(value, Sequence_) and len(value)
         and isinstance(value[0], str) and casefold(value[0]) == casefold(type_str_))
        for type_str_ in type_str)

# ###############
# Recursive reduction

def deep_reduce(v, **kwargs):
    from .functions import serialize_function  # HACK !! We should put this in utils or something, or better – avoid the need to special case Callables at all
    return (v.deep_reduce(v, **kwargs) if isinstance(v, Serializable)
            else Dataclass.deep_reduce(v, **kwargs) if (is_dataclass(v) and not isinstance(v, type))
            else serialize_function(v) if isinstance(v, FunctionType)  # Only serialize plain functions – not callable classes, methods, lambdas, etc.
            else v)

# ###############
# Base class for custom serializable types

# NB: In pydantic.py, we associate the `Serializable.reduce` to the *abstract* base class `ABCSerializable`.
class ABCSerializable(abc.ABC):
    """
    This is the class added to the list of JSON encoders.
    Any type can be registered as a virtual subclass, which will then divert
    its serialization to the hierarchy managed by `Serializable`.

    This abstract class is used by `Serializable` to register parent types
    of the same name, so that they are also serializable.

    (The reason we need a separate class, is that only abstract classes allow
    virtual subclasses, but we don't want to make `Serializable` abstract.)
    """
    # Registry of types which are registered as virtual subclasses of ABCSerializable.
    # This mirrors but differs from Serializable._registry: Both have the same
    # keys, but the values in the latter are true subclasses of Serializable
    # (and so provide `Data`, `reduce`, `__get_validators__`, etc.)
    # It should always be true that an entry in Serializable._registry is a
    # non-strict subclass of the corresponding entry in ABCSerializable._registry.
    # NB: Keys are the *types* (values) of Serializable._registry; this is to
    #     allow subclasses of `Serializable` to redefine _registry to limit the
    #     accepted sub classes.
    _base_types = {}

# NB: Conceptually this is an abstract class, but since it is intended for use
# as a mixin, we don't inherit abc.ABC (the validation functionality is reimplemented
# in `__init_subclass__` instead). Otherwise subclasses would need to implement
# custom metaclasses whenever their parents aren't also ABC subclasses.
@ABCSerializable.register  # Registering Serializable allows any subclass of Serializable to be recognized, without having to register each individually
class Serializable:
    """
    An abstract base class used to standardize the definition of
    Pydantic-serializable types, and reduce the boilerplate to do so.

    A `Serializable` subclass `SubT` must define a nested class `Data`, which
    inherits from `scityping.BaseModel` and defines the data structure used to
    serialize the class. The attribute names of the `Data` class must match the
    keywords of the class' ``__init__``.

    The nested `Data` class must also define a method `encode`, which
    takes an instance of `SubT`, extracts the attributes required for
    serializing it, and returns them as a `SubT.Data` instance.

    .. hint::
       If a `Serializable` has *the same name* (case insensitive) as one of
       its parent classes, then any instance of the *parent* class will be
       serializable. Specifically, the class' `__qualname__` attribute is
       used for this somewhat magical behaviour.

    .. admonition:: Pickle support

       Serializable objects automatically support pickling if their
       `Data.encode` method returns something that is pickleable.
       Note that this only applies to *subclasses* of `Serializable`; many of
       the provided types in this package (including `~scityping.numpy.Array`,
       `~scityping.scipy.Distribution`) *don't* return instances of themselves
       when validating values, instead returning clean, unmodified base types
       (the stated examples return `numpy.ndarray` and `scipy.stats.<rv_frozen>`
       respectively). For these types, pickling if and only if the base types
       supports it.

       Among the types defined in this package,
       `~scityping.functions.PureFunction` and its subclasses is probably the
       one which most benefits from pickling support.
    """
    # Subclasses are stored in a registry
    # This allows deserializers to be retrieved based on the serialized name.
    # Each subclass creates its own _registry, and adds itself to the registry
    # of all its parents
    _registry = TypeRegistry()

    @property
    @classmethod
    @abc.abstractmethod
    def Data(self):
        """
        Any subclass of `Serializable` must define a nested class `Data`,
        which itself defines a class method `encode`.
        """
        # @classmethod
        # @abc.abstractmethod
        # def encode(cls, value) -> Data:
        #     raise NotImplementedError
        raise NotImplementedError

    def __init_subclass__(cls):
        ## Reproduce ABC functionality, since we don't subclass ABC ##
        if not inspect.isabstract(cls):
            if getattr(cls.Data, '__isabstractmethod__', False):
                raise TypeError(f"Can't instantiate class {cls.__qualname__} with "
                                "undefined `Data`.")
            # Ensure that all subclasses define their own `Data` class
            # (Simply inheriting Data from the parent class makes deserialization ambiguous.)
            if (any(cls.Data is getattr(C, "Data", None) for C in type.mro(cls)[1:])
                and not issubclass(cls, Dataclass)):  # Dataclass has special support to allow it to be used as the type for Data without causing infinite recursion
                raise TypeError(f"The `Serializable` type {cls.__qualname__} must "
                                "define its own `Data` class.")
            # Check that 'encode' is a classmethod (https://stackoverflow.com/a/19228282)
            if (not hasattr(cls.Data, "encode")
                  or not isinstance(cls.Data.encode, Callable_) 
                  # or not inspect.ismethod(cls.Data.encode)
                  # or cls.Data.encode.__self__ is not cls.Data  # NB: This does not prevent subclassing Data to reuse the encode
                  ):
                raise TypeError(f"{cls.__qualname__} is a subclass of `Serializable`, and "
                                "therefore its nested `Data` class must define "
                                "a class method `encode`.")
        ## Update subclass registries ##
        # Create a new registry for this class, and add to all the parents
        cls._registry = TypeRegistry()
        for C in type.mro(cls):  # In contrast to `cls.mro()`, works also with metaclasses
            if issubclass(C, Serializable):
                C._registry[cls] = cls
        # Update the registry of base clasess
        # We keep registries of both the base classes (for isinstance checks)
        # and the serializable class (for serialization functions).
        # See /scityping/docs/implementation_logic
        # (Conceptually, one can think of _base_types as a registry of super classes,
        # and _registry as a registry of subclases, perhaps we could get away with iterating through _registry ?)
        ABCSerializable._base_types[cls] = (cls,)
        # MAGIC: Register the lowest class in the MRO OF THE SAME NAME as a virtual subclass.
        #        This is to allow adding serializers for existing data types.
        name = cls.__qualname__.lower()
        try:
            basecls = next(C for C in type.mro(cls)[::-1]
                           if C.__qualname__.lower() == name
                              and not issubclass(C, Serializable))
            # Consider the case here where we define a plain Dataset in `base`, then
            # define `C = Dataset(base.Dataset, torch...Dataset). We want to
            # register the `torch...Dataset` with C. (Any parent which
            # (is Serializable will already have been registered above.)
        except StopIteration:
            pass
        else:
            cls.register(basecls)

    @classmethod
    def register(cls, T:type):
        """Register type `T` as serializable by `cls`."""
        if not issubclass(T, Serializable):
            # Registering everything serializable as virtual subclass would do no harm, but seems unclean (and might cause a performance hit)
            ABCSerializable.register(T)
        # Update _base_types with a minimal set of base types, including `T`
        ABCSerializable._base_types[cls] = tuple(
            base for base in ABCSerializable._base_types[cls]
            if not issubclass(base, T)) + (T,)
        # Tell all the parents of `cls` that they can serialize `T`.
        for C in type.mro(cls):
            if issubclass(C, Serializable) and T not in C._registry:
                C._registry[T] = cls


    @classmethod  # Pydantic validation hook
    def __get_validators__(cls):
        yield cls.validate

    def __reduce__(self):  # Pickle serialization hook
        return (self.validate, (self.deep_reduce(self),))

    @classmethod
    def validate(cls, value, field=None):  # `field` not currently used: only there for consistency
        """
        Validate `value` to an instance of `cls` (possibly of a subclass).
        If it is already of this type, return `value` unchanged.
        Otherwise apply type coercion or deserialization, as required.
        If nothing succeeds, raise `TypeError`.
        For precise details, see /scityping/docs/implementation_logic#validation-logic
        """
        # Branch 1: Check if `value` is already of a serializable type.
        #           If so, simply return `value` unchanged.
        if any(isinstance(value, subcls)
               for subcls in cls._registry.values()):
            return value

        # Branch 2: Check if it matches one of the base types wrapped by serializers
        #           If so, cast it to the corresponding subclass by calling `subclass.validate` on `value`.
        base_types = ABCSerializable._base_types
        matching_basesubclasses = {subcls: set(type.mro(subcls))
                                   for subcls in cls._registry.values()
                                   if isinstance(value, base_types[subcls])}
        if matching_basesubclasses:
            # Remove any subcls which is a strict superclass of an also matching subclass
            to_remove = []
            for subcls, mro in matching_basesubclasses.items():
                if any(mro < othermro for othermro in matching_basesubclasses.values()):
                    to_remove.append(subcls)
            for subcls in to_remove:
                del matching_basesubclasses[subcls]
            assert len(matching_basesubclasses) > 0, "Bug in the logic removing less specific Serializable types."
            if len(matching_basesubclasses) > 1:
                raise NotImplementedError(
                    "Diamond inheritance between Serializable subclasses is not yet "
                    f"supported.\nType {type(value)} subclasses {matching_basesubclasses}.")
            else:
                # Recast value to be of type 'cls'
                subcls = next(iter(matching_basesubclasses))
                if "generator" in str(value).lower():
                    json = subcls.reduce(value)
                    subcls.validate(json)
                return subcls.validate(subcls.reduce(value))  # NB: reduce() doesn’t do the actual conversion to string, so this isn’t too expensive

        # Branch 3: Check if `value` is an instantiated `Data` object.
        #           If so, construct the target type from it.
        types_with_matching_dataclasses = [
            subcls for subcls in set(cls._registry.values())  # We may have duplicate entries (e.g. point both 'Generator' and 'NPGenerator' to NPGenerator)
            if isinstance(subcls.Data, type) and isinstance(value, subcls.Data)]   # The first check is mostly for Data given as an abstractmethod/property in an abstract base class
        if types_with_matching_dataclasses:
            #   `types_with_matching_dataclasses` may have multiple entries when we specialize a type.
            #   For example, if we specialize `Website` into `CommerceWebsite`
            #       class Website(Serializable):
            #         class Data:
            #           ...
            #       class CommerceWebsite(Website):
            #         class Data(Website.Data)
            #           ...
            #   both `Website` and `CommerceWebsite` will be in `BaseWebsite._cls_registry`.
            #   Two different situations can cause multiple matches
            #   1) `CommerceWebsite` does not define a nested `Data` at all and we call
            #          Website.validate(website_data: Website.Data)
            #      This is prohibited by __init_subclass__ because it cannot be resolved.
            #   2) `CommerceWebsite` subclasses `Website.Data`, as in the example above, and we call
            #          Website.validate(website_data: CommerceWebsite.Data)
            #      In this case we want to deserialize to the most specific type: CommerceWebsite
            #   Only when there is multiple inheritance between Serializable types should we need to raise an error.

            # # Remove matching types which have strict subclasses
            # for C in types_with_matching_dataclasses[:]:
            #     if any(issubclass(D, C)
            #            for D in types_with_matching_dataclasses if D is not C):
            #         types_with_matching_dataclasses.remove(C)
            assert len(types_with_matching_dataclasses) == 1, \
                f"Unable to resolve the target data type for value of type {type(value)}; " \
                f"all these Serializable types match: {types_with_matching_dataclasses}." \
                "This can happen if the types’ Data classes have diamond inheritance"
            targetT = types_with_matching_dataclasses[0]
            decoder = getattr(targetT.Data, 'decode', None)
            # If `value` is a plain dataclass, apply our basic support for deserialization
            if (is_dataclass(value) and not isinstance(value, type)
                  and not hasattr(value, "__pydantic_model__")):
                value = validate_dataclass(value, inplace=True)
            # Use either custom or default decoder
            if decoder:
                return decoder(value)
                # Here we skip the call to `validate`: If targetT defines a `decode`,
                # then it expects its value to be used as-is.
                # This allows `decode` to be used to define a different type
                # (e.g. `range` instead of `Range`) without causing an infinite
                # loop (since then the code does not exit on branch #1).
            else:
                # We support dataclass, Pydantic BaseModel and mappings for the nested Data class
                if is_dataclass(value) and not isinstance(value, type):
                    # Make a shallow copy of the fields (`dataclasses.asdict` copies everything, and recurses into dataclasses, dicts, lists and tuples)
                    kwds = {field.name: getattr(value, field.name) for field in fields(value)}
                elif hasattr(value, "__fields__"):
                    # Assume a Pydantic BaseModel
                    kwds = dict(value)
                elif hasattr(value, "items"):
                    kwds = dict(value.items())
                else:
                    raise TypeError(f"Nested `Data` type for {targetT.__qualname__} "
                                    "should be either a dataclass, a Pydantic BaseModel "
                                    f"or a mapping. inheritance for {targetT.__qualname__}:\n"
                                    f"{targetT.Data.mro()}.")
                obj = targetT(**kwds)
                # If targetT has an overridden `validate`, ensure it is executed on the serialized data
                # If `validate` is not overridden, this will exit via branch #1
                return targetT.validate(obj)

        # Branch 4: Check if `value` is a serialized `Data` object.
        #           If so, deserialize it and call `targetT.validate`
        #           (which will use Branch 3 above, unless `validate` is overridden).
        elif (isinstance(value, Sequence_)
              and len(value) == 2
              and isinstance(value[0], str)):
            targetT = cls._registry.get(value[0])
            if targetT is None:
                valuestr = str(value)
                if len(valuestr) > 1000: valuestr = valuestr[:999] + "…"
                raise TypeError("Serialized data does not match any of the registered "
                                f"`Serializable` subclasses of {cls}.\n"
                                f"    Serializable subclasses: {list(cls._registry)}\n"
                                f"    Type received: {type(value).__qualname__}; registry key: '{type(value).__qualname__.lower()}'\n"
                                f"    Value received: {value}\n\n")

            if isinstance(value[1], targetT.Data):  # This can happen if we call `reduce` but never actually convert the result to string
                data = value[1]
            elif isinstance(value[1], Sequence_):
                data = targetT.Data(*value[1])
            elif isinstance(value[1], Mapping_):
                data = targetT.Data(**value[1])
            else:
                raise TypeError(f"Serialized {cls.__qualname__}.Data must be either "
                                "a tuple (positional args) or dict (keyword args). "
                                f"Received {type(value[1])}.")
            return targetT.validate(data)  # Unless overriden is targetT, this will use the `matching_dataclasses` branch above

        # Branch 5: As a final option, try simply passing the value as argument to this type
        #           This always fails on the base Serializer class, since it takes no arguments,
        #           but it allows things like ``PintQuantity.validate(3.)``
        # EDIT: Don't try passing as args, because this can prevent exceptions
        #   from being raised (e.g. an array will accept a heterogeneous list).
        #   Pydantic in particular relies on the "wrong" types raising exceptions,
        #   in order to try the next type in a list.
        else:
            # Give up and print an error message
            valuestr = str(value)
            if len(valuestr) > 1000: valuestr = valuestr[:999] + "…"
            raise TypeError(f"Unable to validate the received value to the type {cls}."
                            f"    Type received: {type(value).__qualname__}; registry key: '{type(value).__qualname__.lower()}'\n"
                            f"    Value received: {valuestr}\n\n")
                            # f"Attempting `{cls}(<value>)` raised the "
                            # f"following error:\n{e}")
    @classmethod
    def reduce(cls, value, **kwargs) -> Tuple[str, "Data"]:
        """
        Generic custom serializer. All serialized objects follow the format
        ``(type name, serialized data)``. For example,
        ``(scityping.range, {'start': 0, 'stop': 5, 'step': none})``.
        The type name is inferred from the `value`’s type with `get_type_key`
        and is case insensitive. It is the key used by `scityping.utils.TypeRegistry`.

        The class used to serialize `value` is inferred by matching the type
        of `value` to one of the entries in `cls._registry`.

        Additional positional and keyword arguments are passed on to
        ``S.Data.encode(*args, **kwargs)``, where ``S`` is the class used to
        serialize `value`. Only `kwargs` which match an argument in the signature
        of `S.Data.encode` are passed on; this allows `kwargs` to contain encoding
        arguments for different encoders, as long as all encoders use different
        argument names. (The use case for this filtering is `deep_reduce`, which
        through recursion may encode many objects but `kwargs` can only be
        specified once.)
        """
        # This first line deals with the case when `reduce` is called with
        # the parent class (typically `Serializable`), or when the serializer
        # is attached to a base class (e.g. the same serializer for all subclasses of np.dtype) 
        for C in type.mro(type(value)):  # In contrast to `type(value).mro()`, this works also when `value` is a type
            serializer_cls = cls._registry.get(C)
            if serializer_cls:
                break
        else:
            # I'm not sure what the most appropriate error type should be, but at least TypeError is consistent with the similar error in `validate()`
            # Also `KeyError` is special and doesn't display newlines; I think it is best reserved for cases where the user interacts directly with a mapping
            raise TypeError("There seems to be no encoder registered for any "
                           f"of the types {type.mro(type(value))}. Note that types "
                           "depending on external librairies (like scipy or torch) "
                           "are only registered when the corresponding module in "
                           "scityping is imported.\n"
                           f"Attempted to serialize using: {cls}.reduce\n"
                           f"Registered types for this class: {cls._registry.keys()}")

        # Call Data.encode, passing any parameters from kwargs that match a parameter in Data.encode.
        if kwargs:
            encode_kwargs = inspect.signature(serializer_cls.Data.encode).parameters.keys()
            kwargs = {kw: arg for kw, arg in kwargs.items() if kw in encode_kwargs}
        data = serializer_cls.Data.encode(value, **kwargs)

        # NB: To reduce boilerplate, we allow `encode` to return a tuple or dict
        # These are respectively treated as positional or keyword args to Data
        if not isinstance(data, serializer_cls.Data):
            if isinstance(data, tuple):
                if issubclass(serializer_cls.Data, BaseModel):
                    # As a convenience, if Data class is a BaseModel, it has a
                    # `__fields__` attribute we can use to construct an argument dictionary
                    fields = serializer_cls.Data.__fields__
                    kwds = {k: v for k, v in zip(fields, data)}
                    data = serializer_cls.Data(**kwds)
                else:
                    data = serializer_cls.Data(*data)

            elif isinstance(data, dict):
                data = serializer_cls.Data(**data)
            else:
                data = serializer_cls.Data(data)
                # raise TypeError(
                #     f"{serializer_cls.__qualname__}.Data.encode method returned an object of "
                #     f"type '{type(data)}'. It must return either an instance of "
                #     f"{serializer_cls.__qualname__}.Data, a tuple or a dict. If a tuple or "
                #     f"dict, these must be arguments to {serializer_cls.__qualname__}.Data.")
        return (get_type_key(serializer_cls), data)

    @classmethod
    def deep_reduce(cls, value, **kwargs) -> Tuple[str, Union[dict,tuple]]:
        """
        Call `reduce` recursively, converting each result to either a tuple
        or a dictionary. The returned value therefore only contains types
        which can be serialized without `scityping`.
        If certain encoders (i.e. the ``Data.encode`` methods) take additional
        parameters, these can be passed as keyword arguments. To avoid
        unexpected values, different encoders should take care to use different
        argument names.
        """
        type_key, data = cls.reduce(value, **kwargs)

        if is_dataclass(data):
            reduced_type = dict
            data_iter = ((k,getattr(data,k)) for k in data.__dataclass_fields__)
        elif isinstance(data, BaseModel):
            reduced_type = dict
            data_iter = data
        elif isinstance(data, Mapping):
            reduced_type = dict
            data_iter = data.items()
        elif isinstance(data, Iterable_):
            reduced_type = tuple
            data_iter = data
        else:
            raise TypeError(f"`{cls.__qualname__}.reduce` returned an object of unexpected "
                            f"type '{type(data).__qualname__}', which is incompatible with "
                            f"`{cls.__qualname__}.deep_reduce`.")

        if reduced_type is dict:
            reduced_data = {k: deep_reduce(v)
                            for k, v in data_iter}
        else:
            reduced_data = tuple(deep_reduce(v) for v in data_iter)

        return (type_key, reduced_data)


class SerializedMeta(type):
    # # Make a singleton
    # __instance = None
    # def __new__(cls):
    #     return SerializedGen.__instance or super().__new__(cls)
    # Actual implementation
    def __getitem__(cls, T: Type[Serializable]) -> Type:
        if cls is not Serialized:
            raise RuntimeError("Type definitions of the form `Serialized[…][…]` "
                               "are not supported.")
        elif not isinstance(T, type):
            raise TypeError("Serialized[…] only accepts types which are "
                            f"subclasses of Serializable. {T} is not even a type.")
        elif not issubclass(T, Serializable):
            raise TypeError("Serialized[…] only accepts types which are "
                            "subclasses of Serializable. "
                            f"{T} does not subclass Serializable.")
        # return Tuple[str, T.Data]
        return SerializedMeta(f"Serialized{T.__name__}", (Serialized,),
                              {"_serialized_type": T, "__slots__": ()})

class Serialized(tuple, metaclass=SerializedMeta):
    """
    Constructor for a type describing serialized objects.
    If ``MyType`` is a subclass of `Serializable`, then ``Serialized[MyType]``
    returns a type which is essentially ``Tuple[str, T.Data]`` but which also:
    - accepts serialized data from subclasses of ``MyType``;
    - accepts instances of ``MyType`` (in which case they are serialized).
    """
    __slots__ = ()
    _serialized_type: ClassVar

    @classmethod
    def __get_validators__(cls):
        yield cls.validate
    @classmethod
    def validate(cls, value):
        # If we got `MyType` instead of `Serialized[MyType]`, convert it to the right form.
        if isinstance(value, cls._serialized_type):
            value = cls.reduce(value)
        # Raise error if value is not of correct type
        if ( not isinstance(value, Sequence_) or len(value) != 2 or not isinstance(value[0], str) ):
            value_str = str(value)
            if len(value_str) > 150: value_str = value_str[:149] + "…"
            raise TypeError(f"A value matching Serialized[{cls._serialized_type.__qualname__}] "
                            "should be a tuple of length two, where the first element is a string. "
                            f"Received:\n {value_str}")
        # Determine the target type for the data by inspecting the first arg.
        # This might not be the same as `_serialized_type`: it could a subclass
        targetT = cls._serialized_type._registry.get(value[0])
        if targetT is None:
            raise TypeError(f"'{value[0]}' does not match any of the subtypes "
                            f"registered to {cls.__qualname__}.")
        # targetT.Data might be a BaseModel, dataclass, TypedDict, etc.
        # Also value[1] might be serialized or not.
        # So we inspect the type and value to determine how to return the tuple (key, Data)
        if isinstance(value[1], targetT.Data):
            # Nothing to do
            return cls(value)
        elif hasattr(targetT.Data, "validate"):
            return cls((value[0], targetT.Data.validate(value[1])))
        elif hasattr(targetT.Data, "__get_validators__"):
            data = value[1]
            for validator in targetT.Data.__get_validators__():
                data = validator(data)
            return cls((value[0], data))
        elif isinstance(value[1], dict):
            return cls((value[0], targetT.Data(**value[1])))
        elif isinstance(value[1], Iterable_):
            return cls((value[0], targetT.Data(*value[1])))

#####################
# Dataclass
#####################
# (Included here instead of base_types to allow import by scityping.json.py)

import builtins
import numbers
import re
from inspect import getmodule
from dataclasses import dataclass, FrozenInstanceError
from typing import Union, Dict, Generic
try:
    from typing import _BaseGenericAlias as BaseGenericAlias
except ImportError:
    from typing import _GenericAlias as BaseGenericAlias  # Python ⩽ 3.8
try:
    from typing import _SpecialForm
except ImportError:  # Python ⩽ 3.7
    class _SpecialForm: pass

# The `Dataclass` type is special cased: it is identified with `is_dataclass`
# rather than with `isinstance`, so we don't need to register a base class.
# (Nor could we, since there is no base dataclass)

dc_val_msg = ("Note that `scityping` provides only basic deserialization "
              "capabilities for plain dataclasses defined in the standard "
              "library. For richer types or to customize the validation, "
              "please use e.g. a Pydantic BaseModel or Pydantic dataclass.")

def split_type_str(s):
    """
    Split on ',', leaving inside of brakets untouched
    "List[int], Dict[str, float]" -> ["List[int]", "Dict[str, float]"]
    """
    els = []
    el_c = []
    counter = 0
    for c in s:
        if c == "[":
            counter += 1
        elif c == "]":
            if counter <= 0:
                raise RuntimeError(f"Unequal brackets in {s}")
            counter -= 1
        if c == "," and counter == 0:
            els.append("".join(el_c))
            el_c = []
        elif c == " ":
            pass  # Drop white space
        else:
            el_c.append(c)
    if counter != 0:
        raise RuntimeError(f"Unequal brackets in {s}")
    els.append("".join(el_c))
    return els

def validate_dataclass_field(val, T: type):
    """
    .. rubric:: Supported types

    Scalar types

    - All subclasses of Serializable
    - int
    - float
    - str
    - bytes
    - bool
    - type (via base_types.Type)
    - tuple
    - dict
    - list
    - set
    - frozenset
    - scityping.NoneType

    Generic types

    - Union
    - Optional
    - Any
    - Tuple
    - List
    - Set
    - FrozenSet
    - Tuple
    - Literal (args must not be Serializable types)
    - Dict
    - Type (via base_types.Type)
    - Callable (accepts all callables; deserializes only PureFunction)

    """

    from . import base_types  # Imported here to avoid cyclic import

    __origin__ = getattr(T, "__origin__", None)

    ## Simplest case: `val` is already of type `T` ##
    # NB: This works even if `T` is not a deserializable type

    # if __origin__ is None and isinstance(T, type) and isinstance(val, T):
    #     # A generic type like Callable[[int],None] is still a type, but would
    #     # raise TypeError in the isinstance (also we want to continue and 
    #     # check the signature). The __origin__ skips generic types.
    #     return val
    try:                          # In Python 3.11+, things like typing.Any are
        if isinstance(val, T):    # now types, but still don’t work with generics.
            return val            # The most reliable seems to be to catch the
    except TypeError:             # TypeError, rather than detect
        pass


    ## Support for generic types ##

    # Union
    if __origin__ is Union or isinstance(T, UnionType):
        if isinstance(val, tuple(_T for _T in T.__args__
                                 if not (isinstance(_T, (BaseGenericAlias, _SpecialForm))
                                         or (isinstance(_T, type) and issubclass(_T, Generic)))  # Some types inherit from Generic but not _BaseGenericAlias (e.g. 'numpy.typing._array_like._SupportsArray')
                                 )):
            # Similar to smart_union: if value is already an instance of any type, don’t change it
            # CAVEAT: Using `isinstance` with generic types like `List[int]` raises TypeError, so we need to remove them from the check.
            #         Consequently this does not work if `val` should match a generic type: then we just do the normal left-to-right coercion. 
            #         `isinstance` also doesn't work with subclasses of _SpecialForm: Any, NoReturn, ClassVar, Union, Optional
            return val
        else:
            # Try coercing, going left to right in the types
            err_msgs = []
            for subT in T.__args__:
                try:
                    return validate_dataclass_field(val, subT)
                except TypeError as e:
                    err_msgs.append(str(e))
            else:
                # None of the types succeeded in deserializing: raise error, with all concatenated messages
                err_str = "\n".join(f"{subT} - {msg}"
                                    for subT, msg in zip(T.__args__, err_msgs))
                err_str = textwrap.indent(err_str, "  ")
                raise TypeError(f"Unable to validate to {T}. "
                                f"Error messages were:\n{err_str}")
    # List, Set, Frozenset
    elif __origin__ in (list, set, frozenset):
        assert len(T.__args__) == 1, \
            f"{'List' if __origin__ is list else 'Set' if __origin__ is set else 'FrozenSet'}[]" \
            " should have only one argument"
        return __origin__(validate_dataclass_field(el, T.__args__[0])
                          for el in val)
    # Tuple
    elif __origin__ is tuple:
        if T.__args__[-1] == ...:
            assert len(T.__args__) == 2, "Tuple[] should have only one type argument when it ends with '...'"
            return tuple(validate_dataclass_field(el, T.__args__[0])
                         for el in val)
        else:
            if len(val) != len(T.__args__):
                raise TypeError(f"Tuple field expects a value with length {len(T.__args__)}, "
                                f"but received one of length {len(val)}.")
            return tuple(validate_dataclass_field(el, elT)
                         for el, elT in zip(val, T.__args__))
    # Literal
    elif __origin__ is Literal:
        if val not in T.__args__:
            raise TypeError(f"Value {val} is not one of the values prescribed by {T}.")
        return val
    # Dict
    elif __origin__ is dict:
        assert len(T.__args__) == 2, "Dict[] should have exactly two arguments"
        keyT, valT = T.__args__
        return {validate_dataclass_field(k, keyT): validate_dataclass_field(v, valT)
                for k, v in val.items()}
    # Type
    elif __origin__ is base_types.Type or __origin__ is type:
        args = getattr(T, "__args__", None)
        if args:
            assert len(args) == 1, "Type[] should have only one argument."
            targetT = base_types.Type[args[0]]
            return targetT.validate(val)
        else:
            # A plain `typing.Type` with no argument will ĥave `__origin__` but no `__args__`
            return base_types.Type.validate(val)
    # Callable
    elif __origin__ is Callable_:
        from .functions import matching_signature, deserialize_function
        if isinstance(val, Callable_):
            # `val` is already a callable: just need to check signature
            if matching_signature(val, T):
                return val
            else:
                raise TypeError(f"Signature of function {val} does not match {T}.")
        elif (isinstance(val, Sequence_)
              and len(val) == 2
              and isinstance(val[0], str)):
            if "PureFunction" in val[0]:
                return Serializable.validate(val)
            else:
                raise TypeError(f"Serialized '{val[0]}' is not a recognized Callable type")
        elif isinstance(val, str):
            logger.warning("Calling `deserialize_function` without any  namespace: this "
                           "is likely to fail except for simple functions. For more reliable "
                           "deserialization, ensure functions are deserialized via `PureFunction`.")
            return deserialize_function(val)
        else:
            raise TypeError(f"{val} is not a Callable")

    ## Support for Any ##

    # Check if it looks like a Serializable type. If so, try to deserialize.
    # Otherwise, or if deserialization fails, return as-is
    # (A tuple of strings e.g. would look like serializable data.)
    if T is Any:
        if (isinstance(val, Sequence_)
              and len(val) == 2
              and isinstance(val[0], str)):
            # TODO: Call directly the right branch, rather than repeat the checks in Serializable.validate
            try:
                return Serializable.validate(val)
            except:
                return val
        else:
            return val

    ## Support for builtin and Serializable types ##

    # From this point, we only support normal types (not those in typing.py)
    if not isinstance(T, type):
        raise TypeError(f"{T} does not support deserialization. {dc_val_msg}")

    # Since we know T is a type, we can use issubclass safely
    if issubclass(T, Serializable):
        return T.validate(val)
    # Subclasses of ABCSerializable don't have a `validate` method, so
    # we just treat them as a non-serializable type.
    # Could we do better ?
    elif is_dataclass(T):
        validate = (getattr(T, "validate", None)
                    or getattr(T, "__validate__", None)  # Used by pydantic.dataclass & scityping.pydantic.dataclass
                    or Dataclass.validate)
        return validate(val)
    elif T is StrictStr:
        if not isinstance(val, str): raise TypeError(f"{val} is not a string")
        return val
    elif T is StrictBytes:
        if isinstance(val, str): val = val.encode()  # Needed to allow deserialization from JSON
        elif not isinstance(val, bytes):  raise TypeError(f"{val} is not a bytes")
        return val
    elif T is StrictBool:
        if isinstance(val, str):
            if val.lower() == "true": val = True
            elif val.lower() == "false": val = False
            else: raise TypeError(f"{val} is not a bool")
        elif not isinstance(val, bool): raise TypeError(f"{val} is not a bool")
        return val
    elif T is StrictInt:
        if isinstance(val, str):
            if re.fullmatch(r"\d+", val): val = int(val)
            else: raise TypeError(f"{val} is not a integer")
        elif not isinstance(val, int): raise TypeError(f"{val} is not a integer")
        return val
    elif T is StrictFloat:
        if isinstance(val, str):
            if re.fullmatch(r"\d*\.\d*", val) and len(val) > 1: val = float(val)
            else: raise TypeError(f"{val} is not a float")
        elif not isinstance(val, float): raise TypeError(f"{val} is not a float")
        return val
    elif issubclass(T, numbers.Integral):
        return int(val)
    elif issubclass(T, numbers.Real):
        return float(val)
    elif issubclass(T, str):
        return str(val)
    elif issubclass(T, bytes):  # Emulate how Pydantic coerces bytes
        if isinstance(val, bytes):
            return val
        elif isinstance(val, str):
            return val.encode()
        elif isinstance(val, numbers.Number):
            return str(val).encode()
        if isinstance(val, bytearray):
            return bytes(val)
        else:
            raise TypeError(f"Coercion of values of type {type(val)} to "
                            "`bytes` is not supported.")
    elif issubclass(T, bool):
        return bool(val)
    elif issubclass(T, base_types.Type):
        T.validate(val)
    elif T is tuple:
        if isinstance(val, tuple):
            return val
        elif isinstance(val, Sequence_):
            return tuple(val)
        else:
            raise TypeError(f"Value {val} is not a tuple.")
    elif T in (list, set, frozenset):
        targetT = T
        if isinstance(val, targetT):
            return val
        elif isinstance(val, Sequence_):
            return TargetT(val)
        else:
            raise TypeError(f"Value {val} is not a {targetT.__name__}.")
    elif T is dict:
        if isinstance(val, dict):
            return val
        elif isinstance(val, Mapping_):
            return dict(val)
        else:
            raise TypeError(f"Value {val} is not a dict.")
    elif T is type(None):
        if val is None:
            return None
        else:
            raise TypeError("A field of type `NoneType` only accepts `None` values. "
                            f"Received {val}")
    else:
        # As a last attempt, check if the type provides a `validate` method.
        # This covers the types in base_types like `NoneType` and Number` which
        # don’t subclass Serializable
        validate = getattr(T, "validate", None)
        if validate:
            return validate(val)
        else:
            raise TypeError(f"Unrecognized type {T}. {dc_val_msg}")

def _get_type_hints(dc):
    """
    A wrapper for typing.get_type_hints which:
    - Merges the namespaces of all classes in the MRO (not just the lowest one)
    - Avoids errors when special values like dataclasses.KW_ONLY are defined
      in a parent class.
    """
    # # Retrieve the namespace in which to search for type definitions
    # dc_mod = getmodule(dc)
    # ns = {**builtins.__dict__, **(dc_mod.__dict__ if dc_mod else {})}
    # HACK: Merge all namespaces for dataclasses in the MRO, so that the types
    #       for fields defined in parent classes are still accessible
    #       Better would be to resolve them with a different namespace for each class in the MRO
    mod_mro = [getmodule(C) for C in type(dc).mro() if is_dataclass(C)]

    # get_type_hints requires a real dict
    ns = dict(ChainMap(*(m.__dict__ for m in mod_mro), vars(builtins)))

    # get_type_hints raises an error if forwardrefs resolve to something which is not a type.
    # Unfortunately, the special value dataclasses.KW_ONLY is not a type => so we need to
    # temporarily replace it with a type.
    # The alternative would be to require that any module which subclasses a dataclass
    # must import all its field types, even hidden ones like KW_ONLY
    try:
        from dataclasses import KW_ONLY
    except ImportError:
        kwonly_placeholder = None  # Python <3.10 does not have KW_ONLY, so nothing to do
    else:
        kwonly_placeholder = type("kwonly_placeholder", (), {})
        for k, v in ns.items():
            if v is KW_ONLY:
                ns[k] = kwonly_placeholder

    ann_types = get_type_hints(type(dc), ns)  # get_type_hints will merge annotations from parents, but only when applied to `type(dc)`

    # Now replace the KW_ONLY placeholder with its true value
    if kwonly_placeholder is not None:
        for C in type(dc).mro():
            for k, v in getattr(dc, "__annotations__", {}).items():
                if v is kwonly_placeholder:
                    dc.__annotations__[k] = KW_ONLY

    return ann_types

def validate_dataclass(dc, inplace=False):
    """
    Args:
        dc: dataclass instance
        inplace: bool
            If `True`, attempt to modify the dataclass `dc` in place.
            If `False`, or `dc` is frozen, a new dataclass is created.

    .. Important:: If `dc` is a frozen dataclass, it cannot be updated in place.
       In such cases a value of `inplace=True` is implicitely converted to
       `inplace=False`. Therefore, one should always assign the return value
       instead of relying on in-place updates.

    .. Note:: Deserialization support for std lib dataclasses is purposefully
       limited to simple cases. Attempting to support every corner would
       quickly turn into a morass of conditionals, and duplicate a lot of the
       functionality of already mature libraries like Pydantic. Essentially
       only JSON types (str, int, float, bool) and `Serializable` types are
       supported.

       One reason we support dataclasses is because we use standard lib 
       dataclasses to serialize our own types is to avoid making Pydantic a
       hard dependency `scityping`. The features we need to do this are those
       which we support for validation.
    """
    ann_types = _get_type_hints(dc)

    new_kwds = {}
    # Validate each field
    for dc_field in fields(dc):
        _val = getattr(dc, dc_field.name, MISSING)
        if _val is MISSING:
            # If some fields have `init=False`, a dataclass can be initialized
            # with some fields unset. In this case there is nothing to validate.
            continue
        dc_field_type = ann_types[dc_field.name]
        # dc_field_type = dc_field.type
        # # NB: field.type is often stored as a string
        # if isinstance(dc_field_type, str):
        #     dc_field_type = get_type_annotation(dc_field_type, ns)
        if isinstance(dc_field_type, str):
            # get_type_annotation was unable to resolve the type annotation
            raise RuntimeError(f"Unable to resolve the type '{dc_field_type}'. "
                               "Are you sure the type is defined "
                               f"in the module '{mod_mro[0].__name__}'?")

        _val = validate_dataclass_field(_val, dc_field_type)

        if inplace:
            try:
                setattr(dc, dc_field.name, _val)
            except FrozenInstanceError:
                inplace = False
                if dc_field.init:
                    new_kwds[dc_field.name] = _val
        elif dc_field.init:  # Discard non-init fields; assume the dataclass will recreate them
            new_kwds[dc_field.name] = _val

    # Return the dataclass if it was modified in place, or create a new one
    if inplace:
        return dc
    else:
        return type(dc)(**new_kwds)

# NB: We don’t decorate with `@dataclass`, so the type can be used for both frozen and mutable dataclasses
class Dataclass(Serializable):
    """
    This class provides very basic serialization/deserialization support for
    standard lib dataclasses.

    Supported types:

    - `int`, `float`, `str`, `bytes`
    - subclasses of `Serializable`
    - Union, tuple, list, set, frozen set, dict

    NOT supported:

    - Pretty much anything else.

    For more powerful deserialization capabilities, it is recommended to use
    full-featured serializable types, like Pydantic's `dataclass` and `BaseModel`,
    or rather their variants in `scityping.pydantic`. Use `dataclass` if you
    just need support for more types, and `BaseModel` if you also need to
    customize how fields are validated.
    Nevertheless, this basic support may be useful

    - as a “first rung” of serialization dependencies: more complex Serializable
      types can use a plain `dataclass` as their `Data` class, as long as they
      limit themselves to the supported types;
    - to avoid adding a dependency on an external package like Pydantic;
    - when dealing with dataclass objects created by an external library, for
      which we cannot change the type.

    .. Note:: Because there is no base class common to all dataclasses,
       dataclasses cannot be identified with the usual way of checking for a
       recognized base type. Their support is hard-coded and specific to
       dataclasses, using `is_dataclass` to recognized them. Since dataclasses
       may be used for the `Data` container, they also use custom code paths
       for deserialization to avoid recursion loops.

    .. Caution:: Use of plain dataclasses as field types for a serializable class
       is also limited. If `MyDataclass` is a plain dataclass type, then the
       following will deserialize as expected

       .. Code::Python
          class MyType(BaseModel):
            dc: MyDataclass

       but the following will not

       .. Code::Python
          class MyType(BaseModel):
            dcs: List[MyDataclass]

       Here again, if you need full serialization/deserialization support, you
       can use `scityping.pydantic.BaseModel` or `scityping.pydantic.dataclass`.

    .. Rubric:: Special cases
       - Fields with ``field(init=False)`` are not serialized. The assumption
         is that these are automatically set by the class (typically in __post_init__),
         and may be internal e.g. caches which we *don’t* want to serialize.
         In cases where they should be included, define a subclass of
         `Dataclass` with its own nested `Data`.

    """

    @classmethod
    def reduce(cls, dc, **kwargs):  # **kwargs required for cooperative signature
        return (get_type_key(cls), cls.Data.encode(dc))
    @classmethod
    def deep_reduce(cls, value, **kwargs):
        type_key, (T, dc_dict) = cls.reduce(value, **kwargs)
        reduced_dict = {k: deep_reduce(v, **kwargs) for k, v in dc_dict.items()}
        return (type_key, (deep_reduce(T, **kwargs), reduced_dict))
    # @classmethod
    # def __get_validators__(cls):
    #     yield cls.validate
    # We need to replace branch 3, since Data doesn't actually do deserialization
    @classmethod
    def validate(cls, value, field=None):
        # Whatever we do, if `value` is a dataclass, the first step is to validate it
        if is_dataclass(value):
            value = validate_dataclass(value, inplace=True)
        # Branch 1: Already a dataclass of the desired type
        if isinstance(value, cls):
            return value
        # Branch 2: `Value` is a Data object. We may first need to validate its content,
        #   which we then use to instantiate `cls` (c.f. Serializable.validate, branch 3)
        elif isinstance(value, cls.Data):
            decoder = getattr(cls.Data, "decode", None)
            # Branch 2a: If cls.Data provides a decoder, use that.
            if decoder:
                return decoder(value)
            # Branch 2b: Subclasses which override Data to set their own fields
            #   (Overriding Data can make serialized data more compact, since it
            #   avoid an extra nesting level with 'type' and 'data' fields.)
            elif (is_dataclass(cls) and is_dataclass(value)  # NB: Dataclass itself is not a dataclass
                  and value.__dataclass_fields__.keys() <= cls.__dataclass_fields__.keys()):
                dc_kwargs = {f.name: getattr(value, f.name) for f in fields(value)}  # Make a shallow dict
                return cls(**dc_kwargs)
            # Branch 2c: Fallback - `cls` provides a `Data`, but not `Data.decode`.
            #   In particular this means they do not subclass `Dataclass.Data`.
            #   The more likely assumption would be that `Data` just includes
            #   dataclass fields, but that situation is covered by branch 2b.
            #   So we try the default decoder, but print a warning, since there
            #   is a reasonable change of this assumption being wrong.
            else:
                logger.error(f"The class {cls.__qualname__} defines a nested `Data` class, "
                             "but this Data class does not define a `decode` method. "
                             "Falling back to the default one provided by `Dataclass.Data`. "
                             f"You can avoid this message by making `{cls.__qualname__}.Data` "
                             f"a subclass of `Dataclass.Data")
                obj = Dataclass.Data.decode(value)
                if ( (cls is not Dataclass and isinstance(obj, cls))   # Normal isinstance()
                     or (cls is Dataclass and is_dataclass(obj)) ):    # When using `Dataclass` as an ABC, any dataclass is valid
                    return obj
                else:
                    raise TypeError(f"Expected a value of type {cls.__qualname__}, "
                                    f"but data serialized to a value of type {type(obj).__qualname__}.")
        # Branch 3: `value` is a generic dataclass. It might still be of the
        #    right type, depending of its `type` field.
        elif isinstance(value, Dataclass.Data):
            obj = Dataclass.Data.decode(value)
            if ( (cls is not Dataclass and isinstance(obj, cls))   # Normal isinstance()
                 or (cls is Dataclass and is_dataclass(obj)) ):    # When using `Dataclass` as an ABC, any dataclass is valid
                return obj
            else:
                raise TypeError(f"Expected a value of type {cls.__qualname__}, "
                                f"but data serialized to a value of type {type(obj).__qualname__}.")
        # Branch 3: `Dataclass` can be used as an ABC for all dataclasses.
        #   In this case we have not specified any fields, so there is also nothing to do.
        elif cls is Dataclass and is_dataclass(value):
            return value
        # Branch 4: None of our dataclass-specific branches matched, so we punt
        #   to the base’s `validate()`.
        else:
            return super().validate(value, field)

    @dataclass
    class Data:
        type: Type
        data: Dict[str, Any]  # TODO: Rename 'kwargs'
        def encode(dc: Dataclass) -> Dataclass.Data:
            # NB: dataclasses.asdict would make a recursive deepcopy of dc, which we don't want
            return (type(dc), {f.name: getattr(dc, f.name)
                               for f in fields(dc) if f.init})
                # Fields which are excluded from init are not serialized.
                # This is necessary with the default decoder, since such fields
                # cannot be given to the __init__ function.
        def decode(data: Dataclass.Data) -> Dataclass:
            obj = data.type(**data.data)
            obj = validate_dataclass(obj, inplace=True)
            return obj



"""
Manifest
========

Types:
------

+ Distributions

  - UniDistribution
  - MvDistribution (deserialization only)
  - MvNormalDistribution

JSON encoders
-------------

+ stats

  - scipy.stats._distn_infrastructure.rv_frozen
  - scipy.stats._multivariate.multi_rv


Extending support to new distribution types
===========================================

For distributions defined in scipy.stats, the best place to add support for
them is in this module. However it is also possible for external packages to
extend the `Distribution` hierarchy; this can be useful in at least two
situations:

* To support custom subclasses of scipy distributions;
* To add support for a missing scipy distribution type, without having to
  maintain one's own version of `scityping`.

The main additional consideration when extending these classes in a separate
module is to ensure that all types used in annotations here are also imported
in the module providing the extension. Otherwise Pydantic will complain about
unprepared types still being a `ForwardRef`::

    from scipy import stats
    from pydantic.dataclasses import dataclass

    from scityping.scipy import Optional, Dict, Tuple, Array,  RVArg, RNGState   # Required to resolve forward refs
    from scityping.scipy import Union, NoneType, Covariance  # Required to resolve forward refs for multivariate distributions
    from scityping.scipy import MvDistribution

    MvT = stats._multivariate.multivariate_t_frozen

    class MultivariateTDistribution(MvDistribution, MvT):
        @dataclass
        class Data(MvDistribution.Data):
            @staticmethod
            def encode(rv, include_rng_state=True):
                ...

"""

# ####
# SciPy statistical distributions
from __future__ import annotations

import abc
import logging
from dataclasses import asdict

import numpy as np
from scipy import stats

from .utils import ModuleList
from .base import json_like, Serializable, ABCSerializable
from .base_types import SerializedData, NoneType, Real

from typing import Optional, Union, Any, Tuple, List, Dict
from .numpy import Array, NPGenerator, RandomState

logger = logging.getLogger(__name__)

# Define types used in Data objects, so classes wanting to extend this know what to import
RVArg = Union[Array[np.number], float, int, Any, None]  # Union of all accepted types for distribution arguments
RNGState = Union[None, NPGenerator, RandomState]
# Scipy.stats does not provide a public name for the frozen dist types, and the
# private paths depend on the version of scipy.
RVFrozen = next(C for C in type(stats.norm()).mro() if "rv_frozen" in C.__name__.lower())  # NB: Ensure we don't use rv_continuous_frozen
MvRVFrozen = next(C for C in type(stats.multivariate_normal()).mro() if "frozen" in C.__name__.lower() and "normal" not in C.__name__.lower())
MvNormalFrozen = next(C for C in type(stats.multivariate_normal()).mro() if "frozen" in C.__name__.lower())
# Similar for the Covariance subclasses, added in SciPy 1.10
if hasattr(stats, "Covariance"):
    stats_has_cov_class = True
    stats_CovViaDiagonal = type(stats.Covariance.from_diagonal(1))
    stats_CovViaCholesky = type(stats.Covariance.from_cholesky(1))
    stats_CovViaEigenDecomposition = type(stats.Covariance.from_eigendecomposition(([1.], [[1.]])))
    stats_CovViaPrecision = type(stats.Covariance.from_precision(1))
    stats_CovViaPSD = stats._covariance.CovViaPSD   # Does not seem to have a public construction API
else:
    stats_has_cov_class = False
# List of modules searched for distribution names; precedence is given to
# modules earlier in the list
# Note that modules can also be specified as strings (so "scipy.stats" would
# also work below). This allows modules to add themselves to this list.
stat_modules = ModuleList([stats])

## Generic definitions ##

class Distribution(Serializable):
    """
    Pydantic-aware type for SciPy _frozen_ distributions. A frozen distribution
    is one for which the parameters (like `loc` and `scale`) are fixed.

    Note that any subclass of `Distribution` must also have its nested `Data`
    be a subclass of `Distribution`. This guarantee facilitates the implementation
    of some decoders, by allowing things like ``isinstance(data, Distribution.Data)``.
    (We use this for example in our Mixture implementation.)

    .. rubric:: Serialization of the random seed

    In order to reproduce the samples produced by statistical distribution, it
    is important to serialize the random state (seed) along with the parameters.
    On the other hand, this state easily represents 95% of the serialized data,
    and transforms a very readable serialization to an unreadable mess.
    So serializing it when only the distribution parameters are needed is undesirable.
    
    To indicate whether random state should be included, use `include_rng_state`
    option of `Data.encode`. As a convenience, the default behaviour is to
    include state only when it was explicitely set.
    """
    class Data(abc.ABC, SerializedData):
        # Some custom distribution types (e.g. mixture) may have distributions,
        # or lists of distributions, as arguments.
        # We accommodate this by including them in the type.
        # NOTE: It is still better to use separate fields for Distribution,
        # and especially List[Distribution] arguments, since those have
        # much more understandable error messages.
        dist: str
        args: Tuple[RVArg, ...]
        kwds: Dict[str, RVArg]
        rng_state: Optional[RNGState]=None

        @staticmethod
        @abc.abstractmethod
        def valid_distname(dist_name: str) -> bool:
            """
            Return `True` if `dist_name` is a possible value for `dist` for
            this `Data` class.
            """
            return False  # Distributions must be serialized with a subclass
        @abc.abstractmethod
        def encode(rv_frozen, include_rng_state: bool=True):  # Implemented by subclasses
            raise NotImplementedError
        def decode(data):
            dist = Distribution.get_dist(data.dist)
            if dist is Distribution:
                raise RuntimeError("The data seem to have been serialized with "
                                   "the generic type `Distribution`; this should not be possible.")
            elif ( isinstance(dist, Serializable)
                   and getattr(dist.Data, "decode", Distribution.Data) is not Distribution.Data ):
                # Second condition intended to prevent infinite recursion
                return dist.Data.decode(data)
            else:
                frozen_dist = dist(*data.args, **data.kwds)
                frozen_dist.random_state = data.rng_state
                return frozen_dist

        def __new__(cls, dist, *a, **kw):
            # Consider if a type annotation specifies Serialized[MvDistribution]
            # and we provide the serialized form of MvNormalDistribution
            # => We would to replace `cls` (set to MvDistribution.Data)
            #    by MvNormalDistribution.Data. This is what we do here:
            #    `data_type` is the `Data` subclass we need to use.
            data_type = Distribution.get_Data_type(dist)
            assert issubclass(data_type, cls), "If `data_type` is not a subclass of `cls`, __init__ will not be called."
            if super().__new__ is object.__new__:
                return super().__new__(data_type)
            else:
                if a:  # There are positional arguments:
                    return super().__new__(data_type, dist, *a, **kw)
                else:  # `dist` might be a keyword-only argument
                    return super().__new__(data_type, dist=dist, **kw)
        # # TODO
        # @classmethod
        # def __modify_schema__(cls, field_schema):
        #     field_schema.update(type='array',
        #                         items=[{'type': 'string'},
        #                                {'type': 'string'},  # dist name
        #                                {'type': 'array'},   # dist args
        #                                {'type': 'array'},   # dist kwds
        #                                {'type': 'array'}    # random state (optional) Accepted: int, old RandomState, new Generator
        #                                ])

    def __init_subclass__(cls):
        data_cls = getattr(cls, "Data", None)
        supercls = next(filter(None,
            (base if issubclass(base, Distribution) else False for base in cls.mro()[1:])))  # The `.Data` attribute of the nearest parent
        if  (  data_cls is None
               or not isinstance(data_cls, type)
               or not issubclass(data_cls, supercls.Data) ):
            base_name = f"{supercls.__module__}.{supercls.__qualname__}"
            raise TypeError(f"{cls.__name__} is a `{base_name}` "
                            "subclass and must define a nested `Data` class "
                            f"as a subclass of `{base_name}.Data`.")
        super().__init_subclass__()

    # The distinction that only frozen dists are serializable is not obvious,
    # so we wrap `validate` to catch that error and print an explanation message
    @classmethod
    def validate(cls, v):
        if isinstance(v, (stats._distn_infrastructure.rv_generic,
                          stats._multivariate.multi_rv_generic)):
            raise TypeError("`Distribution` expects a frozen random variable; "
                            f"received '{v}' with unspecified parameters.")
        return super().validate(v)

    @staticmethod
    def get_dist(name: str):
        """Search the `stat_modules` for a distribution type matching `name`."""
        for module in stat_modules:
            dist = getattr(module, name, None)
            if dist:
                break
        if dist is None:
            raise RuntimeError("Unable to find a distribution named "
                               f"'{name}' within the modules {stat_modules}. "
                               "To add a module to search for distributions, "
                               f"update the list at {__name__}.stat_modules.")
        return dist

    @staticmethod
    def get_Data_type(dist_name: str):
        """
        From the `dist_name` included in serialized data, determine the subclass
        of `Distribution.Data` which produced it.
        """
        data_types = tuple(set(dist_type.Data
                               for dist_type in Distribution._registry.values()
                               if dist_type.Data.valid_distname(dist_name)))
        # Remove types which are strict subclasses of another
        to_remove = set()
        for i, T in enumerate(data_types):
            if issubclass(T, data_types[:i] + data_types[i+1:]):
                to_remove.add(i)
        data_types = tuple(T for i, T in enumerate(data_types) if i not in to_remove)
        # In case of ambiguity, raise error and abort => better fix this error early than let it seep into the code
        if len(data_types) == 0:
            raise RuntimeError("No subclass of `Distribution.Data` seems to recognize "
                               f"the distribution type '{dist_name}'.\nData classes "
                               f"checked: {T.Data for T in Distribution._registry.values()}")
        elif len(data_types) > 1:
            raise RuntimeError(f"Multiple `Data` classes recognize '{dist_name}': "
                               f"{data_types}")
        else:
            return data_types[0]

## Covariance class (SciPy ≥1.10)

if stats_has_cov_class:
    class Covariance(Serializable, stats.Covariance):
        class Data(abc.ABC, SerializedData):
            @abc.abstractmethod
            def encode(cov_object): raise NotImplementedError
            @abc.abstractmethod
            def decode(data): raise NotImplementedError
else:
    # Dummy class so annotations don’t become ForwardRefs
    class Covariance(Serializable):
        class Data(SerializedData):
            def encode(cov): raise NotImplementedError

## Concrete Distribution subclasses ##

class UniDistribution(Distribution, RVFrozen):
    class Data(Distribution.Data):
        @staticmethod
        def valid_distname(dist_name: str) -> bool:
            univariate_names = (o.name for o in stats.__dict__.values()
                                if isinstance(o, stats._distn_infrastructure.rv_generic))
            return dist_name in univariate_names
        def encode(rv, include_rng_state=None):
            """
            include_rng_state: Whether to include the random state when serializing.

            - `True`: Include it.
            - `False`: Do not include it.
            - `None`: Include it iff it differs from the default shared RNG
              used by all distributions unless explicitely set.
            """
            if rv.args:
                logger.warning(
                    "For the most consistent and reliable serialization of "
                    "distributions, consider specifying them using only keyword "
                    f"parameters. Received for distribution {rv.dist.name}:\n"
                    f"Positional args: {rv.args}\nKeyword args: {rv.kwds}")
            if include_rng_state is None:
                # NB: The use of `stats.norm` to access random_state is arbitrary; any standard distribution will work
                include_rng_state = (rv.random_state is not stats.norm.random_state)
            random_state = rv.random_state if include_rng_state else None
            return rv.dist.name, rv.args, rv.kwds, random_state
# NB: We need to special case multivariate distributions, because they follow
#     a different convention (and don't have a standard 'kwds' attribute)
class MvDistribution(Distribution, MvRVFrozen):
    class Data(Distribution.Data):
        @staticmethod
        def valid_distname(dist_name: str) -> bool:
            # Multivariate distributions must always be serialized with subclasses
            return False
        def encode(rv, include_rng_state=None):
            raise NotImplementedError(
                "The reduce for `Distribution` needs to be special "
                "cased for each multivariate distribution, and this has "
                f"not yet been done for '{rv._dist}'.")
class MvNormalDistribution(MvDistribution, MvNormalFrozen):
    class Data(MvDistribution.Data):
        # dist: str
        args: NoneType=None
        kwds: NoneType=None
        # rng_state: Optional[RNGState]=None
        mean: Array[np.inexact, 1]=None                    # Required: default value because dataclass doesn’t allow positional args after optional ones
        cov: Union[Covariance, Array[np.inexact, 2]]=None  # Required: default value because dataclass doesn’t allow positional args after optional ones
        @staticmethod
        def valid_distname(dist_name: str) -> bool:
            return dist_name == "multivariate_normal"
        def encode(rv, include_rng_state=None):
            if include_rng_state is None:
                # NB: The use of `stats.norm` to access random_state is arbitrary; any standard distribution will work
                include_rng_state = (rv.random_state is not stats.norm.random_state)
            dist = rv._dist
            random_state = dist._random_state if include_rng_state else None
            if stats_has_cov_class:  # scipy ≥1.10
                cov = rv.cov_object  # Will be a subclass of stats.Covariance
            else:  # scipy ≤1.9
                cov = rv.cov
            # name, args, kwds, random_state
            return dict(dist="multivariate_normal",
                        mean=rv.mean, cov=cov, rng_state=random_state)
            # return ("multivariate_normal", (),
            #         {'mean':rv.mean, 'cov':cov}, random_state)
        def decode(data):
            mv_normal = stats.multivariate_normal(mean=data.mean, cov=data.cov)
            mv_normal.random_state = data.rng_state
            return mv_normal

UniDistribution.register(RVFrozen)
MvDistribution.register(MvRVFrozen)  # This is only to provide the NotImplementedError message
MvNormalDistribution.register(MvNormalFrozen)

## Concrete Covariance subclasses ##

class _PSD(Serializable, stats._multivariate._PSD):
    # NB: It would be more efficient to store only _M and the __init__ args
    #    (cond, rcond, lower, check_finite and allow_singular)
    #    However I don’t know how to recover those init args from a PSD instance.
    class Data(SerializedData):
        M: Array[np.inexact, 2]  # Don’t use '_M', otherwise Pydantic won’t deserialize
        eps: float
        rank: int
        log_pdet: float
        V: Array[np.inexact, 2]
        U: Array[np.inexact, 2]
        def encode(psd: stats._PSD):
            return (psd._M, psd.eps, psd.rank, psd.log_pdet, psd.V, psd.U)
        def decode(data: "_PSD.Data") -> _PSD:
            # We stored all the attributes: we want to skip __init__ entirely
            # HACK: This may break if the internals of _PSD change
            psd = object.__new__(stats._multivariate._PSD)
            datadict = asdict(data)
            datadict["_M"] = datadict["M"]  # relabel 'M' -> '_M'
            del datadict["M"]
            psd.__dict__.update(**datadict)
            psd._pinv    = None  # Lazily computed attribute
            return psd

if stats_has_cov_class:

    class CovViaPrecision(Covariance, stats_CovViaPrecision):
        class Data(SerializedData):
            precision: Array[Real,2]
            def encode(cov_object): return (cov_object._precision,)
            def decode(data): return stats.Covariance.from_precision(data.precision)

    class CovViaDiagonal(Covariance, stats_CovViaDiagonal):
        class Data(SerializedData):
            diagonal: Array[Real,1]
            def encode(cov_object): return (np.diag(cov_object.covariance),)
            def decode(data): return stats.Covariance.from_diagonal(data.diagonal)

    class CovViaCholesky(Covariance, stats_CovViaCholesky):
        class Data(SerializedData):
            L: Array[Real,2]
            def encode(cov_object): return (cov_object._factor,)
            def decode(data): return stats.Covariance.from_cholesky(data.L)

    class CovViaEigendecomposition(Covariance, stats_CovViaEigenDecomposition):
        class Data(SerializedData):
            eigenvalues: Array[Real, 1]
            eigenvectors: Array[Real, 2]
            def encode(cov_object): return (cov_object._w, cov_object._v)
            def decode(data): return stats.Covariance.from_eigendecomposition(
                (data.eigenvalues, data.eigenvectors))

    class CovViaPSD(Covariance, stats_CovViaPSD):
        class Data(SerializedData):
            psd: _PSD
            def encode(cov_object): return cov_object._psd
            def decode(data): return stats_CovViaPSD(data.psd)




from __future__ import annotations

import logging
from typing import Optional, Union, Dict, List, Tuple
from dataclasses import fields

import holoviews as hv
from .base import Serializable
from .base_types import SerializedData, Number, Type
from .functions import PureFunction
from .typing import StrictStr, StrictInt, StrictFloat

logger = logging.getLogger(__name__)

# Until we support serializing generic parameters, these are the only Parameter
# values that can be serialized
Parameter = Union[StrictStr, StrictInt, StrictFloat]

class Dimension(hv.Dimension, Serializable):
    class Data(SerializedData):
        name        : str
        label       : str
        cyclic      : bool
        default     : Optional[Parameter]
        nodata      : Optional[int]
        range       : Tuple[Union[None,Number], Union[None,Number]]
        soft_range  : Tuple[Union[None,Number], Union[None,Number]]
        step        : Optional[Number]
        type        : Optional[Type]
        unit        : Optional[str]
        value_format: Optional[PureFunction]
        values      : List[Parameter]

        def encode(dim: hv.Dimension) -> Dimension.Data:
            default = dim.default
            values = dim.values
            explain = ("at present a generic serializer for holoviews "
                       "Parameters is not implemented, so the only defaults "
                       f"of type {Parameter.__args__} are serialized.")
            if default is not None and not isinstance(default, Parameter.__args__):
                logger.warning(f"The default value {default} of dimension '{dim.label}' "
                               f"will not be serialized: {explain}")
                default = None
            if dim.values is not None and any(not isinstance(v, Parameter.__args__) for v in dim.values):
                logger.warning("Some or all of the `values` attribute of dimension "
                               f"'{dim.label}' will not be serialized: {explain}")
                values = [v for v in values if isinstance(v, Parameter.__args__)]
                if len(values) == 0:
                    values = None

            return dict(name=dim.name, label=dim.label, cyclic=dim.cyclic,
                        default=default, nodata=dim.nodata, range=dim.range,
                        soft_range=dim.soft_range, step=dim.step, type=type,
                        unit=dim.unit, value_format=dim.value_format,
                        values=dim.values)

        def decode(data: Dimension.Data) -> hv.Dimension:
            # NB: 'name' must be passed as positional arg
            kwargs = {k: getattr(data, k)
                      for k in fields(data) if k != "name"}
            return hv.Dimension(data.name, **kwargs)




import pytest
import numpy as np
from scityping.utils import TypeRegistry

def test_TypeRegistry():
    A = type("A", (), {})
    B = type("B", (), {})

    # Test that `token_sets` attribute is kept in sync
    reg = TypeRegistry(Generator=A)
    reg[np.random.Generator]=A
    reg["torch.Generator"]=B
    assert len(reg._key_tokens) == 3
    del reg["Generator"]
    assert len(reg._key_tokens) == 2
    assert frozenset(("numpy", "random", "_generator", "generator")) in reg._key_tokens

    assert TypeRegistry(Generator=A).get("numpy.random.Generator") is A
    assert TypeRegistry(Generator=A).get(np.random.Generator) is A
    assert TypeRegistry(Generator=A).get("torch.Generator") is A
    assert TypeRegistry({"numpy.Generator":A}).get("torch.Generator") is A
    assert TypeRegistry({"numpy.Generator":A,
                         "torch.Generator":B}).get("torch.Generator") is B
    assert TypeRegistry({"Generator":A,
                         "torch.Generator":B}).get("torch.Generator") is B
    assert TypeRegistry({"torch.Generator":B,                          # Test for logic errors that would make result
                         "Generator":A}).get("torch.Generator") is B   # dependent on the order of registry entries 
    assert TypeRegistry(Generator=A).get("mypkg.Generator") is A

    # Errors
    # TODO: Caplog, error: replaced key
    # with pytest.raises(ValueError):  # Duplicate keys
    #     TypeRegistry(Generator=A, generator=A)

    with pytest.raises(ValueError):  # Ambiguous match
        TypeRegistry({"Generator":A,
                      "torch.Generator":B})["mypkg.Generator"]
    with pytest.raises(KeyError):  # No matching entry
        TypeRegistry(Juniper=A)[complex]
    with pytest.raises(KeyError):
        TypeRegistry({"numpy.Generator": A})["Generator.numpy"]

    with pytest.raises(KeyError):  # The last token must match
        TypeRegistry({"numpy.Generator": A})["numpy"]


from typing import Any
from collections.abc import Callable
from dataclasses import is_dataclass
from .base import ABCSerializable, Serializable, Dataclass
from .functions import serialize_function

try:
    from pydantic.main import ValidationError
except ModuleNotFoundError:
    # If pydantic can’t be imported, it is fine to create a dummy class which
    # will never match
    class ValidationError(RuntimeError):
        pass

# Based on pydantic.json.custom_pydantic_encoder
def scityping_encoder(obj: Any, base_encoder=None) -> Any:
    """
    This function can be used as the `default` argument to `json.dump`, to
    add support for `Serializable` types.
    """
    if isinstance(obj, ABCSerializable):
        try:
            # NB: The function pydantic.json.pydantic_encoder checks if the argument
            #     is a BaseModel or dataclass, and if so, calls respectively .dict()
            #     or asdict() to convert it to a dictionary.
            #     *These are recursive calls*, which means that they reduce
            #     every value inside them using only that function’s machinery:
            #     nested calls to `pydantic_encoder` are *not* made, and in
            #     particular `asdict` will ignore any custom encoders – it even
            #     calls `deep_copy` on its contents.
            # We emulate pydantic..pydantic_encoder here and call a recursive
            # function to also reduce any nested Serializable values within `obj`.
            # If we only reduce the `obj` but not its attributes,
            # pydantic..pydantic_encoder may bypass `Serializable.reduce` for
            # Serializable attributes.
            return Serializable.deep_reduce(obj)
        except ValidationError as e:
            # Pydantic's ValidationError uses obj.__name__ it its error message.
            # With our SciTyping pattern, this results in all error message
            # displaying `Data` as the object, which isn't very useful.
            # To improve the error message, we modify the model attached to
            # the exception so that its __name__ is actually __qualname__.
            # This monkey patching isn't especially clean, but it should be
            # innocuous: after all, we've already aborted code execution.
            if e.model.__name__ == "Data":
                e.model.__name__ = e.model.__qualname__
            raise e
    elif is_dataclass(obj) and not isinstance(obj, type):
        # Dataclasses need to be special cased because there is no
        # base dataclass, so we can't identify them with `isinstance`
        # They are worth special-casing because they are one of the basic types
        # used for packaging data, in particular for our nested Data classes
        return Dataclass.deep_reduce(obj)
    elif isinstance(obj, Callable):        # NB: We don't want to do this within 'Data.encode', because sometimes we
        try:
            return serialize_function(obj)     #     use 'encode' without going all the way to a JSON string
        except TypeError as e:
            # Things like callable Pydantic models may not be serializable with the function serializer,
            # but still serializable with the base encoder.
            # (NB: We don’t try the base encoder first, because we want to give
            # the function serializer priority.)
            if base_encoder:
                return base_encoder(obj)
            else:
                raise e
    elif base_encoder:
        return base_encoder(obj)
    else:
        raise TypeError




import pytest
import logging
import scityping.scipy as stsp
from scityping.pydantic import BaseModel

# TODO: Move tests from statGLOW.stats.tests.test_serialization to here

def test_distribution():
    # Distribution enforces that subclasses also subclass Distribution.Data
    with pytest.raises(TypeError):
        class MyDist(stsp.Distribution):
            class Data:
                dist: str
                args: tuple
                kwds: dict

def test_distributions(caplog):
    from scipy import stats
    from scityping.scipy import (Distribution, UniDistribution,
                                 MvDistribution, MvNormalDistribution)

    for D in [stats.poisson(3.1), stats.norm(-1, scale=3), stats.zipf(a=2.2)]:
        D.random_state = 123
        # By default the RNG is serialized, so both D and D2 return the same values
        with caplog.at_level(logging.ERROR, logger="scityping.scipy"):  # Hide warnings about serializing arguments
            D2 = Distribution.validate(UniDistribution.reduce(D))
        assert (D.rvs(25) == D2.rvs(25)).all()
        # The RNG state is generally the largest amount of data (especially for MT19337),
        # so we can skip serializing if we don't need it.
        with caplog.at_level(logging.ERROR, logger="scityping.scipy"):  # Hide warnings about serializing arguments
            D3 = Distribution.validate(UniDistribution.reduce(D, include_rng_state=False))
        assert (D.rvs(25) != D3.rvs(25)).any()  # Now the generated random numers are different

    # For multivariate distributions, in contrast to univariate ones, a different
    # subclass of `Distribution` is needed for each one.
    D = stats.multivariate_normal([2, -2], [[2.2, 1.],[1., 1.3]])
    D.random_state = 321
    D2 = Distribution.validate(MvNormalDistribution.reduce(D))
    assert (D.rvs(25) == D2.rvs(25)).all()
    D3 = Distribution.validate(MvNormalDistribution.reduce(D, include_rng_state=False))
    assert (D.rvs(5) != D3.rvs(5)).all()  # With R-valued distributions we can be more aggressive in the test and use .all() instead of .any()

    # Not all multivariate distributions are implemented yet
    D = stats.dirichlet([.2, .7])
    with pytest.raises(NotImplementedError):
        MvDistribution.reduce(D)

    # Test complete serialization to JSON
    class Foo(BaseModel):
        dist: Distribution
    for D in [stats.poisson(3.1), stats.norm(-1, scale=3), stats.zipf(a=2.2)]:
        foo = Foo(dist=D)
        foo2 = Foo.parse_raw(foo.json())





import logging
import dataclasses
import pytest

import numpy as np
import pint
from pydantic import validator
from pydantic import ValidationError
from scityping.pydantic import BaseModel, dataclass

from typing import List, Tuple
from scityping.pint import PintUnit, PintValue

def test_pint():
    # NB: Pint v0.18 changed to 'ApplicationRegistry'; before was 'LazyRegistry'
    #     The second `getattr` is in case LazyRegistry is removed in the future
    ApplicationRegistry = getattr(pint, 'ApplicationRegistry',
                                  getattr(pint, 'LazyRegistry', None))

    # PintUnitMeta
    assert PintUnit._ureg is None
    ureg = PintUnit.ureg
    assert isinstance(ureg, ApplicationRegistry)
    assert PintUnit._ureg is ureg
    assert PintUnit.ureg is ureg
    with pytest.raises(TypeError):
        PintUnit.ureg = 3
    PintUnit.ureg = ureg

    # PintUnit
    assert PintUnit.validate(ureg.m) == ureg.m
    assert PintUnit.validate(1.*ureg.m) == ureg.m  # Also accept Quantities, iff they have magnitude 1
    with pytest.raises(ValueError):
        PintUnit.validate(2.*ureg.m)
    u = ureg.m
    u_json = PintUnit.reduce(u)
    assert u_json == ('scityping.pint.PintUnit', PintUnit.Data(name='meter'))
    assert PintUnit.validate(u_json) == ureg.m

    uu = PintUnit(u)
    assert uu.apply(5) == 5*ureg.m == 5*uu
    assert uu.apply(10*ureg.m) == 10*ureg.m == uu.apply(10*uu) == 10*uu
    v = uu.apply(1000*ureg.cm)  # Compatible units get convert
    assert v.units == ureg.m
    assert v == 10*ureg.m
    with pytest.raises(ValueError):
        uu.apply(5*ureg.s)  # `apply` ensures the result has exactly the prescribed units
                            # (in contrast to multiplication, which would simply add those units)

    # PintValue
    assert PintValue.validate(3.*ureg.s) == 3.*ureg.s
    # with pytest.raises(TypeError):
    #     PintValue.validate(3.)
    assert PintValue.validate(3.*ureg.dimensionless) == PintValue(3.) == 3.*ureg.dimensionless
    v = 3.*ureg.s
    vms = v.to('ms')
    v_json = PintValue.reduce(v)
    vms_json = PintValue.reduce(vms)
    assert v_json == ('scityping.pint.PintQuantity', PintValue.Data(data=(3.0, (('second', 1),))))
    assert v_json != vms_json  # Serialization changes depending on units
    assert PintValue.validate(v_json) == PintValue.validate(vms_json)  # But deserialized values still compare equal



from __future__ import annotations

import sys
import importlib
import inspect
import operator
import functools
import re
from collections.abc import Callable as Callable_

from typing import ClassVar, Union, Type, Callable, List, Tuple, _Final
from types import new_class

import numpy  # TODO: Make numpy import optional

from .base import Serializable
from .base_types import SerializedData, Number
from .numpy import Array  # TODO: Make numpy import optional

PlainArg = Union[Number, str, Array]

__all__ = ["PlainArg", "PureFunction", "PartialPureFunction", "CompositePureFunction"]

# Functions from these modules are not serialized by storing their source,
# but simply the name is saved. There are a few reasons to do this:
# - C functions, like the builtins, cannot be serialized by inspecting their source.
# - NumPy ufuncs similarly cannot be serialized by inspection
# - NumPy array functions technically could, but because of the overloading the
#   dependencies are quite complicated.
# - Serializing just the name is much more compact, and arguably more reliably
#   for mature functions like those in core Python or NumPy
pure_function_set = {fn: module.__name__
                     for module in [operator, numpy]
                     for fn in module.__dict__.values()
                     if isinstance(fn, Callable_)
                     }


# TODO: Instead of "trust_all_inputs", use a whitelist of safe module (as in scityping.scipy.Distribution)
# TODO: Serializing partial(PureFunction(g)) should be the same as PartialPureFunction(partial(g)), for consistent hashes
# FIXME!: Specifying types to a PureFunction multiple times causes the following error:
#       The Serializable TypeRegistry already contained an entry for key
#       scityping.functions.purefunctionmeta.__getitem__.<locals>.update_namespace.<locals>.data,
#       which has been replaced.

class PureFunctionMeta(type):
    _instantiated_types = {}
    def __getitem__(cls, args):
        """
        Returns a subclass of `PureFunction`. Args may consist of
        - The callable type (in the same format as `~typing.Callable`).
        - Module names. These are used to define namespaces which should be
          imported into the local namespace during deserialization.
        - Both a callable type and module names.
        
        .. Note::
           Types cannot be specified as strings – string arguments are assumed
           to be module names.
        """
        # Parse the arguments
        callableT = {'inT': None, 'outT': None}
        modules = []
        for a in args:
            if isinstance(a, str):
                modules.append(a)
            elif inspect.ismodule(a):
                for nm, m in sys.modules.items():
                    if m is a:
                        modules.append(nm)
                        break
                else:
                    raise AssertionError(f"Module {a} not found in `sys.modules`.")
            elif isinstance(a, list):
                if callableT['inT'] is not None:
                    raise TypeError("Only one input type argument may be specified to"
                                     f"`PureFunction`. Received {callableT['inT']} and {a}.")
                callableT['inT'] = a
            elif isinstance(a, (_Final, type)) or a is None:
                if callableT['outT'] is not None:
                    raise TypeError("Only one output type argument may be specified to"
                                     f"`PureFunction`. Received {callableT} and {a}.")
                if a is None:
                    a = type(None)  # This is what Callable does automatically anyway, and it allows us to check below that either both of inT, outT were passed, or neither
                callableT['outT'] = a
            else:
                raise TypeError("Arguments to the `PureFunction` type can "
                                "consist of zero or one type and zero or more "
                                f"module names. Received {a}, which is of type "
                                f"type {type(a)}.")
        # Treat the callable type, if present
        if (callableT['inT'] is None) != (callableT['outT'] is None):
            raise TypeError("Either both the input and output type of a "
                            "PureFunction must be specified, or neither.")
        if callableT['inT']:
            assert callableT['outT'] is not None
            baseT = Callable[callableT['inT'], callableT['outT']]
            argstr = f"{callableT['inT']}, {callableT['outT']}"
            __args__ = (*callableT['inT'], callableT['outT'])
                # Not sure that merging inT and outT is the best, but 
                # typing.Callable does it like this
        else:
            baseT = Callable
            argstr = ""
            __args__ = ()
        # Treat the module names, if present
        if modules:
            if argstr:
                argstr += ", "
            argstr += ", ".join(modules)
        # Check if this PureFunction has already been created, and if not, do so
        key = (cls, baseT, tuple(modules))
        # We set Serializable's subtype registry key ourselves, otherwise the 
        # automatically determined one looks like scityping.functions.purefunctionmeta.__getitem__.<locals>.update_namespace.<locals>.data
        # which is not informative and causes name conflicts.
        if callableT['inT']:
            serialization_key = f"{cls.__qualname__}[[{','.join(T.__qualname__ for T in callableT['inT'])}], {callableT['outT'].__qualname__}]"
        else:
            serialization_key = cls.__qualname__  # Equivalent to letting utils.get_type_key automatically determine the key
        if key not in cls._instantiated_types:
            def update_namespace(namespace):
                # All subclasses of Serializable need their own nested Data class
                class Data(cls.Data):
                    __serialization_key__ = serialization_key + ".Data"
                namespace["modules"] = cls.modules + modules
                namespace["Data"] = Data
                namespace["__args__"] = __args__
                namespace["__serialization_key__"] = serialization_key
                namespace["__module__"] = cls.__module__
            PureFunctionSubtype = new_class(
                f'{cls.__name__}[{argstr}]', (cls,),
                exec_body=update_namespace)
            cls._instantiated_types[key] = PureFunctionSubtype
            # To also allow an instance of an untyped PureFunction as a value for
            # a typed field, we need to add PureFunction to the subclass' _registry
            # (See docs/implementation_logic.md::Validation of functions)
            PureFunctionSubtype.register(cls)
        # Return the PureFunction type
        return cls._instantiated_types[key]

class PureFunction(Serializable, metaclass=PureFunctionMeta):
    """
    A Pydantic-compatible function type, which supports deserialization.
    A “pure function” is one with no side-effects, and which is entirely
    determined by its inputs.

    Accepts also partial functions, in which case an instance of the subclass
    `PartialPureFunction` is returned.

    .. Warning:: Deserializing functions is necessarily fragile, since there
       is no way of guaranteeing that they are truly pure.
       When using a `PureFunction` type, always take extra care that the inputs
       are sane.

    .. Note:: Functions are deserialized without the scope in which they
       were created.
       
       To provide additional scope, one or more modules name can be added to the
       list `PureFunction.modules`. These modules are imported deserialization
       but are not serialized with the function.
       Adding modules this way is considered an experimental feature.


    .. Hint:: If ``f`` is meant to be a `PureFunction`, but defined as::

       >>> import math
       >>> def f(x):
       >>>   return math.sqrt(x)

       then it has dependency on ``math`` which is outside its scope, and is
       thus impure. It can be made pure by putting the import inside the
       function::

       >>> def f(x):
       >>>   import math
       >>>   return math.sqrt(x)

    .. Note:: Like `Callable`, `PureFunction` allows to specify the type
       within brackets: ``PureFunction[[arg types], return y]``. However the
       returned type doesn't support type-checking.

    .. Hint:: `PureFunction` instances can be pickled, using the same serializer
       as `Serializable.reduce`. This can make them more portable:
       whereas a pickled python function requires that the original function &
       module still exist at the same location, a pickled `PureFunction`
       deserializes from the source code in the serialized data.
    """
    __args__: ClassVar[Tuple[Type,...]] = ()
    modules: List[str] = []  # Use this to list modules that should be imported into
                             # the global namespace before deserializing the function
    # Instance variable
    func: Callable

    class Data(SerializedData):
        func: Union[Callable]
        def __post_init__(self):
            cls = PureFunction  # TODO? Get the container of type(self) ?
            if not isinstance(self.func, Callable_):
                modules = [importlib.import_module(m_name) for m_name in cls.modules]
                global_ns = {k:v for m in modules
                                 for k,v in m.__dict__.items()}
                # Since decorators are serialized with the function, we should at
                # least make the decorators in this module available.
                local_ns = {'PureFunction': PureFunction,
                            'PartialPureFunction': PartialPureFunction,
                            'CompositePureFunction': CompositePureFunction}
                self.func = deserialize_function(self.func, global_ns, local_ns)
            super().__post_init__()

        def encode(purefunc: PureFunction) -> str:
            return purefunc.func,

    def __new__(cls, func=None):
        # func=None allowed to not break __reduce__ (due to metaclass)
        # – inside a __reduce__, it's fine because __reduce__ will fill __dict__ after creating the empty object
        if isinstance(func, functools.partial) and not issubclass(cls, PartialPureFunction):
            # Redirect to PartialPureFunction constructor
            # FIXME: What if we get here from CompositePureFunction.__new__
            try:
                Partial = cls.__dict__['Partial']
            except KeyError:
                raise TypeError(f"{func} is a partial function but '{cls}' "
                                "does not define a partial variant.")
            else:
                return Partial(func)
        return super().__new__(cls)
    def __init__(self, func):
        if hasattr(self, 'func'):
            # This is our second pass through __init__, probably b/c of __new__redirect
            # assert hasattr(self, '__signature__')
            return
        if isinstance(func, str):
            func = deserialize_function(func)
        self.func = func
        # Copy attributes like __name__, __module__, ...
        functools.update_wrapper(self, func)
        try:
            self.__signature__ = inspect.signature(func)
        except ValueError:  # Some functions, like numpy.sin, don’t support `signature`
            pass
    
    @classmethod
    def validate(cls, value, field=None):  # `field` not currently used: only there for consistency
        # Extra check: Make sure the wrapped function is pure
        # (validation = deserialization, so the function should already be marked pure)
        if isinstance(value, Callable_):
            if not is_pure_function(value):
                raise TypeError(f"Cannot validate {value} as a PureFunction, "
                                "because callables are assumed non-pure by default. "
                                "To indicate that a function is pure, use the "
                                "`scityping.functions.PureFunction` decorator.")
        # Replace / Augment branch 2: If we use a PureFunction to validate a
        # field with type PureFunction[[int], None], we don’t want this to fail
        # outright
        # !! WIP !! With more work, we could be both more stringent (allow less
        # errors) and less stringent (allow things which do fit the pattern.)
            if not matching_signature(value, cls):
                raise TypeError(f"The signature of {value} does not match "
                                f"that required by the type {cls}.\n"
                                "Note that signature checking is still primitive: "
                                "for example, it does not match `int` to "
                                "`Union[int,float]`to bypass the check, use "
                                "`PureFunction` without specifying a signature")
        # Additional shorthand specification format like: "x,y -> do_something"
        elif isinstance(value, str):                           # Instead of checking that string looks valid, let `deserialize_function` fail if necessary
            value = PureFunction(deserialize_function(value))  # This should produce more useful error messages.
        # Continue with the normal validation
        return super().validate(value, field)
    
    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __str__(self):
        return f"{type(self).__name__}({self.func})"

    ## Function arithmetic ##
    def __abs__(self):
        return CompositePureFunction(operator.abs, self)
    def __neg__(self):
        return CompositePureFunction(operator.neg, self)
    def __pos__(self):
        return CompositePureFunction(operator.pos, self)
    def __add__(self, other):
        if other == 0:  # Allows using sum([PureFn]) without creating unnecessary Composite functions
            return self
        return CompositePureFunction(operator.add, self, other)
    def __radd__(self, other):
        if other == 0:  # Idem
            return self
        return CompositePureFunction(operator.add, other, self)
    def __sub__(self, other):
        if other == 0:  # Idem
            return self
        return CompositePureFunction(operator.sub, self, other)
    def __rsub__(self, other):
        return CompositePureFunction(operator.sub, other, self)
    def __mul__(self, other):
        return CompositePureFunction(operator.mul, self, other)
    def __rmul__(self, other):
        return CompositePureFunction(operator.mul, other, self)
    def __truediv__(self, other):
        return CompositePureFunction(operator.truediv, self, other)
    def __rtruediv__(self, other):
        return CompositePureFunction(operator.truediv, other, self)
    def __pow__(self, other):
        return CompositePureFunction(operator.pow, self, other)

    ## Serialization / deserialization ##
    # The attribute '__func_src__', if it exists,
    # is required for deserialization. This attribute is added by
    # `deserialize_function` when it deserializes a function string.
    # We want it to be attached to the underlying function, to be sure
    # the serializer can find it
    @property
    def __func_src__(self):
        return self.func.__func_src__
    @__func_src__.setter
    def __func_src__(self, value):
        self.func.__func_src__ = value

    # TODO: Add arg so PureFunction subtype can be specified in error message
    @classmethod
    def raise_validation_error(cls, value):
        raise TypeError("PureFunction can be instantiated from either "
                        "a callable, "
                        "a Sequence `([PureFunction subtype name], func, bound_values)`, "
                        "or a string. "
                        f"Received {value} (type: {type(value)}).")

# # If pydantic is available, it is used for the Data classes, and we need to resolve the circular dependency
# pydantic_model = getattr(PureFunction.Data, "__pydantic_model__", None)
# if pydantic_model:
#     pydantic_model.update_forward_refs()
# del pydantic_model

class PartialPureFunction(PureFunction):
    """
    A `PartialPureFunction` is a function which, once made partial by binding
    the given arguments, is pure (it has no side-effects).
    The original function may be impure.
    """
    class Data(SerializedData):
        func: Union[Callable]
        args: tuple
        kwargs: dict

        @classmethod
        def encode(cls, purefunc: Union[PartialPureFunction,functools.partial]) -> str:
            if isinstance(purefunc, functools.partial):
                # We need this branch because we register `functools.partial` to PartialPureFunction
                # We want to allow encoding only pure functions, so the wrapped function must be pure
                partialfunc = purefunc
                if isinstance(partialfunc.func, functools.partial):
                    raise NotImplementedError("`PartialPureFunction.Data` does not "
                                              "support nested partial functions at this time")
                elif isinstance(partialfunc.func, PureFunction):
                    # Wrap the partial with PartialPureFunction, and re-apply `encode`
                    return cls.encode(PartialPureFunction(purefunc))
                else:
                    raise TypeError("`PartialPureFunction.Data.encode` can only serialize pure functions")

            # From this point `purefunc` is know to be a PartialPureFunction
            func = purefunc.func
            if isinstance(func, PureFunction):
                return (func, {})  # If `func` wraps a partial, its keywords will be taken care of at a deeper level
            elif not isinstance(func, functools.partial):
                # Serialize the function and store an empty dict of bound arguments
                return (func, (), {})
                # # Make a partial with empty dict of bound arguments
                # func = functools.partial(func)
            elif isinstance(func.func, functools.partial):
                raise NotImplementedError("`PartialPureFunction.Data` does not "
                                          "support nested partial functions at this time")
            else:
                # At this point we know that `func` is a partial
                if isinstance(func.func, PureFunction):
                    return (func.func, func.args, func.keywords)
                else:
                    return (func.func, func.args, func.keywords)

        def decode(data: PartialPureFunction.Data) -> PureFunction:
            return PartialPureFunction(functools.partial(data.func, *data.args, **data.kwargs))

    def __str__(self):
        func, args, kwargs = self.Data.encode(self)
        return f"{type(self).__name__}({func}, {args}, {kwargs})"

PureFunction.Partial = PartialPureFunction
# Allow serializing something like partial(purefunc, a=2)
PartialPureFunction.register(functools.partial)

class CompositePureFunction(PureFunction):
    """
    A lazy operation composed of an operation (+,-,*,/) and one or more terms,
    at least one of which is a PureFunction.
    Non-pure functions are not allowed as arguments.

    Typically obtained after performing operations on PureFunctions:

    >>> f = PureFunction(…)
    >>> g = PureFunction(…)
    >>> h = f + g
    >>> isinstance(h, CompositePureFunction)
    True

    .. important:: Function arithmetic must only be done between functions
       with the same signature. This is NOT checked at present, although it
       may be in the future.
    """
    class Data(SerializedData):
        opname: str
        terms: List[Union[PureFunction, PlainArg]]  # NB: PureFunction first, so it catches str arguments

        def encode(purefunc: CompositePureFunction) -> str:
            assert purefunc.func in operator.__dict__.values(), "CompositePureFunction only supports operations found in the `operator` module."
            return (purefunc.func.__name__, purefunc.terms)
        def decode(data: CompositePureFunction.Data) -> PureFunction:
            cls = CompositePureFunction
            func = getattr(operator, data.opname)
            return cls(func, *data.terms)

    def __new__(cls, func=None, *terms):
        return super().__new__(cls, func)
    def __init__(self, func, *terms):
        if func not in operator.__dict__.values():
            raise TypeError("CompositePureFunctions can only be created with "
                            "functions defined in " "the 'operator' module.")
        for t in terms:
            if isinstance(t, Callable) and not isinstance(t, PureFunction):
                raise TypeError("CompositePureFunction can only compose "
                                "constants and other PureFunctions. Invalid "
                                f"argument: {t}.")
        self.func = func
        self.terms = terms
        if not getattr(self, '__name__', None):
            self.__name__ = "composite_pure_function"

    def __str__(self):
        return f"{type(self).__name__}({self.func}, {self.terms})"

    # TODO? Use overloading (e.g. functools.singledispatch) to avoid conditionals ?
    def __call__(self, *args):
        return self.func(*(t(*args) if isinstance(t, Callable) else t
                           for t in self.terms))


#################################################
##               Utilities                     ##
#################################################

import logging
logger = logging.getLogger(__name__)
from warnings import warn
import builtins
import operator
import inspect
from itertools import chain

from types import FunctionType
from typing import Callable, Optional, Any

import ast
if not hasattr(ast, "unparse"): # unparse is available only for >3.9, but the astunparse package backports to earlier versions
    try:
        import astunparse
    except ModuleNotFoundError:
        logger.warning("You may want to install `astunparse` for more reliable "
                       "function serialization, or upgrade to Python ≥3.9.")
    else:
        ast.unparse = astunparse.unparse
import textwrap

from .config import config
from .utils import UnsafeDeserializationError

def is_pure_function(f: Union[Callable, Any]) -> bool:
    """
    Return True if `f` is (very probably)[#]_ a pure function.
    Whether a function is pure is very difficult to determine automatically,
    and we rely on users to identify which functions are pure. Specifically,
    this returns `True` if one of following conditions is met:

    - `f` is not callable (trivial case)
    - `f` was decorated with `PureFunction`.
    - `f` is defined in one of the following modules, which are know to
      define pure functions:
        + `operator`
        + `numpy` (global namespace only)
    - `f` is a partial function, and original function is pure.
      This is checked by calling `is_pure_function` recursively.
      (corner case: we also check bound arguments, in case on of them is a function)

    In all other situations, the returned value will be False.

    .. [#] We say “very probably” because it is possible, for example, to
       decorate a non-pure function with `PureFunction`, or to monkey patch a
       non-pure function into `operator`. In practice, as long as the user does
       not engage in such self-destructive behaviour, and is careful to only
       mark as pure functions which really are pure, this function is reliable.
    """
    return (not isinstance(f, Callable_)
            or isinstance(f, PureFunction)
            or f in pure_function_set
            or (isinstance(f, functools.partial)
                and is_pure_function(f.func)
                and all(is_pure_function(a) for a in f.args)
                and all(is_pure_function(kw) for kw in f.keywords))
            )

def matching_signature(f: Callable, CallableT: Type):
    """
    Return `True` if the signature of `f` matches the type of `CallableT`.
    `CallableT` would normally be created as either `Callable[[int, int] bool]`
    or `PureFunction[[int, int] bool]`.

    .. CAUTION:: This function currently only supports exact type matches in
       the signature. So if the type is `Callable[[Union[int,float]], bool]`
       and the given function is `def f(a:int) -> bool`, it will fail
       becaues `int == Union[int,float]` returns `False`.
       However, if the function omits type hints for certain arguments, then
       those arguments always match.
    """
    if not getattr(CallableT, "__args__", None):
        return True  # If no types are given, everything matches

    inT = CallableT.__args__[:-1]
    outT = CallableT.__args__[-1]
    sig = inspect.signature(f)
    pos_only_params = {nm: p for nm, p in sig.parameters.items()
                       if p.kind == p.POSITIONAL_ONLY}
    pos_or_kw_params = {nm: p for nm, p in sig.parameters.items()
                        if p.kind == p.POSITIONAL_OR_KEYWORD}
    kw_only_params = {nm: p for nm, p in sig.parameters.items()
                      if p.kind == p.KEYWORD_ONLY}
    req_params = {nm: p for nm, p in sig.parameters.items()
                  #if p.kind in {p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD}
                  if p.default is inspect._empty}
    # TODO? Do something with VAR_POSITIONAL argument ?

    # Check that the number of args match
    if (len(inT) < len(req_params)
          or len(inT) > len(pos_only_params) + len(pos_or_kw_params) + len(kw_only_params)):
        return False

    # If available, check that the output type matches
    if sig.return_annotation is not inspect._empty:
        if sig.return_annotation != outT:
            return False

    # Check that the positional only arguments types match (fixed order)
    for p, T in zip(pos_only_params.values(), inT):
        if p.annotation is inspect._empty:
            continue
        elif p.annotation != T:
            return False

    # Check that the other arguments match.
    # Since they can be passed by keyword, the order does not matter
    # First we check that all required args in the signature match a type in CallableT
    nonpos_inT = list(inT[len(pos_only_params):])
    req_Ts = [p.annotation for p in sig.parameters.values()
              if (p.kind in {p.POSITIONAL_OR_KEYWORD, p.KEYWORD_ONLY}
                  and p.default is inspect._empty
                  and p.annotation is not inspect._empty)]
    opt_Ts = [p.annotation for p in sig.parameters.values()
              if (p.kind in {p.POSITIONAL_OR_KEYWORD, p.KEYWORD_ONLY}
                  and p.default is not inspect._empty
                  and p.annotation is not inspect._empty)]
    no_ann = [p for p in sig.parameters.values()
              if (p.kind in {p.POSITIONAL_OR_KEYWORD, p.KEYWORD_ONLY}
                  and p.annotation is inspect._empty)]
    for sigT in req_Ts:
        try:
            nonpos_inT.remove(sigT)
        except ValueError:
            # The function has a required arg which doesn’t match a type in CallableT
            return False

    # Any remaining types in CallableT must match one in the signature
    # Exception: if an argument has no type annotation, it can match any arg in CallableT
    missed_matches = 0
    for annT in nonpos_inT:
        try:
            opt_Ts.remove(annT)
        except ValueError:
            # CallableT has an arg which doesn’t match one in signature, even an optional one
            missed_matches += 1
    # Number of missed matches must not exceed number of non-annotated arguments in signature
    if missed_matches > len(no_ann):
        return False

    # Passed all the tests => return True
    return True

def split_decorators(s):
    s = s.strip()
    decorator_lines = []
    while s[0] == "@":
        line, s = s.split("\n", 1)
        decorator_lines.append(line)
        s = s.lstrip()
        # print(line); print(decorator_lines); print(s)
    return decorator_lines, s

def remove_comments(s, on_fail='warn'):
    """
    Remove comments and doc strings from Python source code passed as string.
    Based on https://stackoverflow.com/a/56285204

    This function will fail if the string `s` is not valid Python code.

    By default, if an error is raised, the string `s` is
    is returned unchanged and a warning is printed. This follows from the
    idea that `remove_comments` is a 'nice to have' feature, and we would
    rather still save / operator on an unchanged `s` than terminate.

    :param on_fail: Default: 'warn'. Change to 'raise' to raise the error
        instead of simply printing a warning.
    """
    # # The bit below is based on a more recent answer to the same question: https://stackoverflow.com/a/76593305
    # # It might be more robust, but since I don’t know ast (and even less astor), I would need to test more thoroughly
    # try:
    #     parsed = ast.parse(s)
    # except Exception as e:
    #     if on_fail == 'warn':
    #         warn(f"{str(e)}\n`remove_comments` encountered the error above. "
    #              "The string was returned unmodified.")
    #         return s
    #     else:
    #         raise e
    # for node in ast.walk(parsed):
    #     if isinstance(node, ast.Expr) and isinstance(node.value, ast.Str):
    #         # set value to empty string
    #         node.value = ast.Constant(value='') 
    # formatted_code = ast.unparse(parsed)  
    # pattern = r'^.*"""""".*$' # remove empty """"""
    # formatted_code = re.sub(pattern, '', formatted_code, flags=re.MULTILINE) 
    # return formatted_code

    try:
        src = ast.unparse(ast.parse(textwrap.dedent(s)))
    except Exception as e:
        if on_fail == 'warn':
            warn(f"{str(e)}\n`remove_comments` encountered the error above. "
                 "The string was returned unmodified.")
            return s
        else:
            raise e
    # Although this looks different than the quoted answer, the logic is the
    # same: remove lines which start with '"'
    # However we also identify multiline docstrings, something the original code missed
    # Remove multiline strings
    src = re.sub(r'^\s*"""[\s\S]*?"""[ \t]*\n?', "", src, flags=re.MULTILINE)
    src = re.sub(r"^\s*'''[\s\S]*?'''[ \t]*\n?", "", src, flags=re.MULTILINE)
    # Remove single line strings  (must do this after multiline strings !)
    src = re.sub(r'^\s*".*\n', "", src, flags=re.MULTILINE)
    src = re.sub(r"^\s*'.*\n", "", src, flags=re.MULTILINE)
    return src

def serialize_function(f) -> str:
    """
    WIP. Encode a function into a string.
    Accepts only definitions of the form::

        def func_name():
            do_something

    or::

        @decorator
        def func_name():
            do_something

    or functions defined in `operator`::

        import operator
        operator.add

    This excludes, e.g. lambdas and dynamic definitions like ``decorator(func_name)``.
    However there can be multiple `@` decorators.

    Upon deserialization, the string is executed in place with :func:`exec`, and the
    user is responsible for ensuring any names referred to within the function
    body are available in the deserializer's scope.

    .. Note:: Instances of `PureFunction` are not supported. Instead use
       `Serializable.reduce(pure_function)`, which will take care of
       extracting the function and calling `serialize_function` on it.
    """

    if f in pure_function_set:
        # Special case serializing of standard pure functions
        return f"{pure_function_set[f]}.{f.__name__}"  # Only support __name__: In `deserialize_function`, we split on the last period to get the function name
    # if f in operator.__dict__.values():
    #     "Special case serializing builtin functions"
    #     return f"operator.{f.__name__}"
    elif hasattr(f, '__func_src__'):
        return f.__func_src__
    elif isinstance(f, PureFunction):
        raise TypeError("To serialize a `PureFunction`, use `Serializable.reduce(purefunc)`.")
        # return f.serialize(f)  # (Partial/Composite)PureFunction takes care of extracting its `func` attribute and calling `serialize_function` on it
    elif isinstance(f, FunctionType):
        s = remove_comments(inspect.getsource(f))
        decorator_lines, s = split_decorators(s)
        if not s.startswith("def "):
            raise ValueError(
                f"Cannot serialize the following function:\n{s}\n"
                "It should be a standard function defined in a file; lambda "
                "expressions are not accepted.")
        return "\n".join(chain(decorator_lines, [s]))
    else:
        raise TypeError(f"Type {type(f)} is not recognized as a "
                        "serializable function.")

def deserialize_function(
    s: str, globals: Optional[dict]=None, locals: Optional[dict]=None):
    """
    WIP. Decode a function from a string.
    Accepts strings of the following forms::

        def func_name(x, y):
            do_something

    or::

        @decorator
        def func_name(x, y):
            do_something
            
    or::
    
        lambda x,y: do_something
        
    or::
    
        x,y -> do_something

    (The last form is equivalent to a lambda function, and is provided as a
    convenience shorthand.)
    Not accepted are dynamic definitions like ``decorator(func_name)``.
    However there can be multiple decorators using the '@' syntax.

    The string is executed in place with :func:`exec`, and the arguments
    `globals` and `locals` can be used to pass defined names.
    The two optional arguments are passed on to `exec`; the deserialized
    function is injected into `locals` if passed, otherwised into `global`.

    .. note:: The `locals` namespace will not be available within the function.
       So while `locals` may be used to define decorators, generally `globals`
       is the namespace to use.

    .. note:: A few namespaces are added automatically to globals; by default,
       these are ``__builtins__``, ``np`` and ``math``. This can
       be changed by modifying the module variable
       `~scityping.config.default_namespace`.

    .. note:: Both `locals` and `globals` will be mutated by the call (in
       particular, the namespaces mentioned above are added to `globals` if not
       already present). If this is not desired, consider making a shallow copy
       of the dicts before passing to `deserialize_function`.
       
    .. note:: While _de_serialization of lambda functions is possible,
       serializing them is not currently supported in general. (This is because
       the output of `inspect.getsource` depends on the context, and we want to
       avoid fragile heuristics.)
       _Re_serialization of a lambda function is possible however.
    """
    msg = ("Cannot decode serialized function. It should be a string as "
           f"returned by inspect.getsource().\nReceived value:\n{s}")
    # First check if this is a special-cased function; if so, exit early
    if isinstance(s, str) and " " not in s:  # Any function def should have white space, and no identifier can have white space
        modname, fname = s.rsplit(".", 1)       
        if modname not in sys.modules:
            raise RuntimeError(f"Cannot deserialize function '{s}': "
                               f"module {modname} is not imported.")
        return getattr(sys.modules[modname], fname)
    # if isinstance(s, str) and s.startswith("operator."):
    #     return getattr(operator, s[9:])
    # Not a builtin: must deserialize string
    if not config.trust_all_inputs:
        raise UnsafeDeserializationError(
            "Deserialization of functions saved as source code requires executing "
            "them with `exec`, and is only attempted if "
            "`scityping.config.trust_all_inputs` is set to `True`.")
    if globals is None and locals is not None:
        # `exec` only takes positional arguments, and this combination is not possible
        raise ValueError("[deserialize]: Passing `locals` argument requires "
                         "also passing `globals`.")
    if isinstance(s, str):
        s_orig = s
        if "def " in s:
            f = _deserialize_def(s, globals, locals)
        elif "lambda " in s or s.count("->") == 1:
            f = _deserialize_lambda(s, globals, locals)
        else:
            raise ValueError(msg)
        # Store the source with the function, so it can be serialized again
        f.__func_src__ = s_orig
        return f
    else:
        raise ValueError(msg)

def _deserialize_def(s, globals, locals):
    decorator_lines, s = split_decorators(s)
    if not s[:4] == "def ":
        msg = ("Cannot decode serialized function. It should be a string as "
               f"returned by inspect.getsource().\nReceived value:\n{s}")
        raise ValueError(msg)
    fname = s[4:s.index('(')].strip() # Remove 'def', anything after the first '(', and then any extra whitespace
    s = "\n".join(chain(decorator_lines, [s]))
    if globals is None:
        globals = config.default_namespace.copy()
        exec(s, globals)
        f = globals[fname]
    elif locals is None:
        globals = {**config.default_namespace, **globals}
        exec(s, globals)
        f = globals[fname]
    else:
        globals = {**config.default_namespace, **globals}
        exec(s, globals, locals)  # Adds the function to `locals` dict
        f = locals[fname]
    return f
    
def _deserialize_lambda(s, globals, locals):
    # `s` is either of the form `lambda x,…: <expr>` or `x,… -> <expr>`
    s = s.strip()
    if not s.startswith("lambda"):
        assert s.count("->") == 1
        inp, out = s.split("->")
        s = f"lambda {inp}: {out}"
    else:
        assert "->" not in s
    if globals is None:
        globals = config.default_namespace.copy()
    else:
        globals = {**config.default_namespace, **globals}
    return eval(s, globals, locals)



"""
Pydantic-aware classes for Pint objects.

In most cases, this should use the correct unit registry automatically.
If that is not the case, the registry can be set by assigning it as a
*class* attribute to `PintUnit`:
>>> import pint
>>> ureg = pint.UnitRegistry()
>>>
>>> from scityping.pint import PintUnit
>>> PintUnit.ureg = ureg
"""

import abc
from typing import Union, Any, Tuple
import pint

from .base import ABCSerializable, Serializable
from .base_types import SerializedData, Number

UnitRegistry = pint.UnitRegistry
ApplicationRegistry = getattr(pint, 'ApplicationRegistry',
                              getattr(pint, 'LazyRegistry', None))
# NB: Pint v0.18 changed to 'ApplicationRegistry'; before it was 'LazyRegistry'.
#     The second `getattr` is in case LazyRegistry is removed in the future

class PintQuantity(Serializable, pint.Quantity):  # Using the same name => a serializers for pint.Quantity is automatically added
    """
    A value with pint units, e.g. `1*ureg.s`.
    """
    class Data(SerializedData):
        data: Tuple[Any, Tuple[Tuple[str, Number],...]]  # Value, units
        def encode(v): return {'data': v.to_tuple()}
        def decode(data): return PintUnit.ureg.Quantity.from_tuple(data.data)

PintValue = PintQuantity  # Deprecated; for backwards compatibility
Quantity = PintQuantity   # To match the name in Pint

type_to_register = pint.Quantity
cls = PintQuantity
ABCSerializable.register(type_to_register)
ABCSerializable._base_types[cls] = type_to_register
for C in cls.mro():
    if issubclass(C, Serializable):
        C._registry[type_to_register] = cls

class PintUnitMeta(abc.ABCMeta):  # Inherit from ABCMeta because PintUnit subclasses Serializable

    @property
    def ureg(cls) -> Union[UnitRegistry, ApplicationRegistry]:
        return cls._ureg or cls._get_and_set_application_registry()

    @ureg.setter
    def ureg(cls, value: Union[UnitRegistry, ApplicationRegistry]):
        if not isinstance(value, (UnitRegistry, ApplicationRegistry)):
            raise TypeError(f"The registry assigned to `{cls.__name__}.ureg` must be a "
                            f"Pint UnitRegistry. Received {value} (type: {type(value)}).")
        cls._ureg = value
    def _get_and_set_application_registry(cls):
        cls._ureg = pint.get_application_registry()
        return cls._ureg

class PintUnit(Serializable, pint.Unit, metaclass=PintUnitMeta):
    # TODO: Type check assignment to ureg
    _ureg: UnitRegistry = None

    def apply(self, value) -> pint.Quantity:
        """
        Give `value` these units. In contrast to using multiplication by a unit
        to construct a `~Pint.Quantity`, `apply` is meant for situations where
        we are unsure whether `value` already has the desired units.

        - If `value` has no units: equivalent to multiplying by `self`.
        - If `value` has the same units as `self`: return `value` unchanged.
        - If `value` has different units to `self`: raise `ValueError`.

        Perhaps a more natural, but more verbose, function name would be
        "ensure_units".
        """
        if (not isinstance(value, pint.Quantity)
              or value.units == PintUnit.ureg.dimensionless):
            return value * self
        elif value.units == self:
            return value
        else:
            try:
                return value.to(self)
            except pint.DimensionalityError as e:
                raise ValueError(f"Cannot apply units `{self}` to value `{value}`: "
                                 f"it already has units `{value.units}`.") from e

    class Data(SerializedData):
        name: str
        def encode(unit): return str(unit)
        def decode(data): return PintUnit.ureg.Unit(data.name)

    # Allow converting a magnitude 1 Quantity to a Unit
    @classmethod
    def validate(cls, v, field=None):
        if isinstance(v, pint.Quantity):
            if v.magnitude != 1:
                raise ValueError("Quantities can only be converted to units "
                                 "if they have unit magnitude.")
            return v.units
        else:
            return super().validate(v, field=field)

type_to_register = pint.Unit
cls = PintUnit
ABCSerializable.register(type_to_register)
ABCSerializable._base_types[cls] = type_to_register
for C in cls.mro():
    if issubclass(C, Serializable):
        C._registry[cls] = cls
        C._registry[type_to_register] = cls



"""
Type annotations and serializers for NumPy types
"""

from __future__ import annotations

import abc
import logging
from types import SimpleNamespace
from typing import (Optional, Union, Any, Literal,
                    Iterable, Sequence, Tuple, List, Dict)
from dataclasses import asdict
from functools import lru_cache

from .base import json_like, Serializable, ABCSerializable, validate_dataclass
from .base_types import SerializedData
from .utils import LazyDict, get_type_key

try:
    from pydantic import StrictInt, StrictFloat
except ModuleNotFoundError:
    StrictInt, StrictFloat = int, float

import numpy as np

import io
# import blosc
import base64

cache = lru_cache(maxsize=None)  # Compatibility with ≤3.8
logger = logging.getLogger(__name__)

# ##################
# Helpers for NumPy types

def get_dtype(T: Any) -> np.dtype:
    """
    Provide a NumPy type object; `T` may be a plain Python type, a NumPy type
    or dtype, or a type which inherits from the `NPValue` type defined in this
    module.

    Returns
    -------
    type
        Returns a NumPy dtype.
    """
    if isinstance(T, type) and issubclass(T, np.generic):
        # It's important not to call `np.dtype(T)` if T is already a NumPy
        # type, since it prevents generics (e.g. `np.dtype(np.number) is np.float64`)
        # It also triggers a deprecation warning for this reason.
        # T = T.dtype
        pass
    else:
        T = np.dtype(T).type
    return np.dtype(T)

# ###############
# DType


@ABCSerializable.register  # NB: It is not permitted to subclass `np.dtype`
class DType(Serializable):
    """
    Validator also accepts `str` and scalar types (int, np.float32, etc.).
    """
    class Data(SerializedData):
        desc: str
        def encode(dtype): return str(dtype)
        def decode(data): return np.dtype(data.desc)  # Required since we don’t subclass `np.dtype`

    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     field_schema.update(type="str")

    @classmethod
    def validate(cls, value, field=None):
        # Following works with any NumPy, but also plain types like int or str
        if isinstance(value, type) and np.issubdtype(value, np.generic):
            return np.dtype(value)
        else:
            return super().validate(value, field=field)

DType.register(np.dtype)
# Because the name 'DType' doesn't match the class we want to make serializable
# (in fact we want to make many classes serializable), we update the registry manually
for type_name in ('int8', 'int16', 'int32', 'int64',
                  'uint8', 'uint16', 'uint32', 'uint64',
                  'float16', 'float32', 'float64', 'float128',
                  'complex64', 'complex128', 'complex256',
                  'bool_', 'str_'):  # Same list as for NPType
    try:
        dtype = np.dtype(type_name)
    except TypeError:
        # Not all machines define all types; e.g. float128 is not always defined
        pass
    else:
        DType.register(type(dtype))
#         T = type(dtype)
#         if T not in DType._registry:
#             DType.register(T)

# ###############
# NPValue type

def infer_numpy_type_to_cast(nptype, value):
    """
    This function tries to determine which concrete numpy type to use to cast
    `value`, given the desired type `nptype`.
    The challenge is that `nptype` may be an abstract type, so we
    have to work through the possible hierarchy until we find the
    most appropriate concrete type. We only do the most common cases;
    we can extend later to the complete tree.
    """
    # If `nptype` is an NPValue subclass, get the true numpy type
    nptype = getattr(nptype, 'nptype', nptype)
    assert issubclass(nptype, np.generic), f"{nptype} is not a subclass of `np.generic`"
    if nptype is np.generic:
        raise NotImplementedError("Unclear how we should cast np.generic.")
    if issubclass(nptype, np.flexible):  # void, str, unicode
        # Assume concrete type
        return nptype
    elif not issubclass(nptype, np.number):  # bool, object
        assert (nptype is np.bool_ or nptype is np.object_), f"Expected NumPy type 'bool' or 'object'; received '{nptype}'"
        return nptype
    else:  # Number
        if issubclass(nptype, np.integer):
            if nptype is np.integer:
                return np.int_
            else:
                # Assume concrete type
                return nptype
        elif issubclass(nptype, np.inexact):
            if issubclass(nptype, np.complexfloating):
                if nptype is np.complexfloating:
                    return np.complex_
                else:
                    # Assume concrete type
                    return nptype
            elif issubclass(nptype, np.floating):
                if nptype is np.floating:
                    return np.float_
                else:
                    # Assume concrete type
                    return nptype
            else:
                assert nptype is np.inexact, f"Expected NumPy type 'inexact'; received '{nptype}'"
                # We try to guess which type to use between float and complex.
                # We make a rudimentary check for complex, and fall back to float.
                if isinstance(value, complex):
                    return np.complex_
                if isinstance(value, str) and ('j' in value or 'i' in value):
                    return np.complex_
                else:
                    return np.float_
        else:
            assert nptype is np.number, f"Expected NumPy type 'number'; received '{nptype}'"
            # We try to guess which type to use between int, float and complex.
            # We make a rudimentary check for int, complex, and fall back to float.
            if isinstance(value, int):
                return np.int_
            elif isinstance(value, complex):
                return np.complex_
            elif isinstance(value, str):
                if 'j' in value or 'i' in value:
                    return np.complex_
                elif '.' in value:
                    return np.float_
                else:
                    return np.int_
            else:
                return np.float_

def convert_nptype(T):
    """
    Convert a type to a NumPy type (e.g. `float` -> `np.float64`).
    If `T` is already a NumPy type, is it returned unchanged.
    """
    if isinstance(T, type) and issubclass(T, np.generic):
        # It's important not to call `np.dtype(T)` if T is already a NumPy
        # type, since it prevents generics (e.g. `np.dtype(np.number) is np.float64`)
        # It also triggers a deprecation warning for this reason.
        pass
    else:
        T = np.dtype(T).type
    return T

class _NPValueType(Serializable, np.generic):
    nptype = None

    class Data(SerializedData):
        data: Union[float, int, str]
        def encode(val): return val.item()

    # There should be no need for the auto-registration in __init_subclass__,
    # and it can add a lot of cruft to the _registries.
    def __init_subclass__(cls):
        pass

    @classmethod
    def validate(cls, value, field=None):
        # NOTE:
        # We make an exception to always allow casting from a string.
        # This allows for complex values, which may not be converted to numbers
        # by the JSON deserializer and still be represented as strings
        if field is None:
            field = SimpleNamespace(name="")
        if isinstance(value, np.ndarray):
            # Allow scalar arrays
            if value.ndim==0:
                value = value[()]
            else:
                raise ValueError(f"Field {field.name} expects a scalar, not "
                                 f"an array.\nProvided value: {value}.")
        # Don't cast unless necessary
        # np.issubdtype allows specifying abstract dtypes like 'number', 'floating'
        # np.generic ensures isubdtype doesn't let through non-numpy types
        # (like 'float'), or objects which wrap numpy types (like 'ndarray').
        if (isinstance(value, np.generic)
            and np.issubdtype(type(value), cls.nptype)):
            return value
        elif (np.can_cast(value, cls.nptype)
              or np.issubdtype(getattr(value, 'dtype', np.dtype('O')), np.dtype(str))):
            # Exception for strings, as stated above
            nptype = infer_numpy_type_to_cast(cls.nptype, value)
            return nptype(value)
        else:
            raise TypeError(f"Cannot safely cast '{field.name}' type  "
                            f"({type(value)}) to type {cls.nptype}.")
    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     if np.issubdtype(cls.nptype, np.integer):
    #         field_schema.update(type="integer")
    #     else:
    #         field_schema.update(type="number")

class _NPValueMeta(type):
    @cache  # Memoization ensures types are created only once
    def __getitem__(self, nptype):
        nptype=get_dtype(nptype).type
        nptype_str = nptype.__name__
        return type(f'NPValue[{nptype_str}]', (_NPValueType,),
                    {'nptype': nptype})

class NPValue(_NPValueType, metaclass=_NPValueMeta):
    """
    Use this to use a NumPy dtype for type annotation; `pydantic` will
    recognize the type and execute appropriate validation/parsing.

    This may become obsolete, or need to be updated, when NumPy officially
    supports type hints (see https://github.com/numpy/numpy-stubs).

    - `NPValue[T]` specifies an object to be casted with dtype `T`. Any
       expression for which `np.dtype(T)` is valid is accepted.

    .. Note:: **Difference with `DType`.** The annotation `NPValue[np.int8]`
       matches any value of the same type as would be returned by `np.int8`.
       `DType` describes an instance of `dtype` and would match `np.dtype('int8')`,
       but also `np.dtype(float)`, etc.

    Example
    -------
    >>> from pydantic.dataclasses import dataclass
    >>> from scityping.numpy import NPValue
    >>>
    >>> @dataclass
    >>> class Model:
    >>>     x: NPValue[np.float64]
    >>>     y: NPValue[np.int8]

    """
    pass

# Because the name 'NPValue' doesn't match the class we want to make serializable
# (in fact we want to make many classes serializable), we update the registry manually
_NPValueType.register(NPValue)  # Required because we removed __init_subclass__
for type_name in ('int8', 'int16', 'int32', 'int64',
                  'uint8', 'uint16', 'uint32', 'uint64',
                  'float16', 'float32', 'float64', 'float128',
                  'complex64', 'complex128', 'complex256',
                  'bool_', 'str_'):  # Same list as for DType
    try:
        dtype = np.dtype(type_name)
    except TypeError:
        # Not all machines define all types; e.g. float128 is not always defined
        pass
    else:       
        _NPValueType.register(getattr(np, type_name))

# ###
# Array type
#
# Arrays are serialized in two different ways, based on whether it is worth
# compressing them vs simply storing them as a string.

# TODO: Expose an interface for configuring compressors/decompressors
# TODO: Allow different serialization methods:
#       - str(A.tolist())
#       - base64(blosc) (with configurable keywords for blosc)
#       - base64(zlib)
#       - external file

encoders = {"b85": base64.b85encode}
decoders = {"b85": base64.b85decode}
# We use LazyDict to avoid forcing an import of a (possibly not installed) module
compressors = LazyDict(none = lambda s: s,
                       blosc= "blosc.compress",
                       zlib = "zlib.compress")
decompressors = LazyDict(none = lambda s: s, 
                         blosc= "blosc.decompress",
                         zlib = "zlib.decompress")
assert encoders.keys() == decoders.keys(), "Each encoder must have an associated decoder."
assert compressors.keys() == decompressors.keys(), "Each compressor must have an associated decompressor."

# Short arrays we just save as string (more legible and more space efficient).
# Since the string loses type information, we save the dtype as well.
class ListArrayData(SerializedData):
    data : Union[list, StrictFloat, StrictInt]  # Scalar types are to deal with 0-dim arrays
    dtype: DType
    @classmethod
    def encode(cls, array):
        return cls(array.tolist(), array.dtype)
    def decode(data):
        return np.array(data.data, dtype=data.dtype)

# Longer arrays are compressed and converted to base85 encoding, with a short summary
# Arrays of size 100 are around the break-even point for 64-bit floats, blosc, base85
_EncoderType = Literal[tuple(encoders)]         # Workaround because the ability
_CompressionType = Literal[tuple(compressors)]  #   of scityping.Dataclass to resolve
class CompressedArrayData(SerializedData):      #   deferred types is limited
    encoding: _EncoderType
    compression: _CompressionType
    summary: str
    data: bytes
    @classmethod
    def encode(cls, array, compression: Union[str,Tuple[str,...]]=("blosc",),
               encoding="b85", threshold: int=100):
        """
        :param:compression:
           A string matching one of the keys in `scityping.numpy.compressors`.
           May also be a tuple of keys; later keys are used as fallback values,
           if earlier entries are not found (e.g. if blosc is not installed.)
        """
        # Parse arguments
        if compression is None:
            compression = 'none'
        if isinstance(compression, (tuple, list)):
            for comp in compression:
                compressor = compressors.get(comp)
                if compressor:
                    compression = comp  # Value of `compression` is stored with the data
                    break
            else:
                raise ModuleNotFoundError("None of the specified compressors "
                                          f"were found: {compression}.")
        else:
            compressor = compressors[compression]
        encoder = encoders[encoding]
        # Convert array to bytes
        with io.BytesIO() as f:  # Use file object to keep bytes in memory
            np.save(f, array)        # Convert array to plateform-independent bytes  (`tobytes` is not meant for storage)
            v_bytes = f.getvalue()
        # Compress and encode the bytes
        array_encoded = encoder(compressor(v_bytes))
        # Set print threshold to ensure str returns a summary
        with np.printoptions(threshold=threshold):
            array_sum = str(array)
        return cls(encoding, compression, array_sum, array_encoded)
    def decode(data):
        decoder = decoders[data.encoding]
        decompressor = decompressors[data.compression]
        v_bytes = decompressor(decoder(data.data))
        with io.BytesIO(v_bytes) as f:
            decoded_array = np.load(f)
        return decoded_array

class _ArrayType(Serializable, np.ndarray):
    nptype = None   # This must be a type (np.int32, not np.dtype('int32'))
    _ndim = None

    # Dispatches between ListArrayData and CompressedArrayData according to array size
    class Data(SerializedData):
        data: Union[ListArrayData, CompressedArrayData]
        # Class variables; Not yet exposed parameters
        compression = ("blosc", "zlib")  # Fall back on zlib compression if blosc is not available
        encoding = "b85"
        threshold = 100
        @classmethod
        def encode(cls, array, compression=None, encoding=None, threshold=None):
            if array.size <= cls.threshold:
                return {"data": ListArrayData.encode(array)}
            else:
                return {"data": CompressedArrayData.encode(
                    array, compression or cls.compression, encoding or cls.encoding,
                    threshold if threshold is not None else cls.threshold)}
        @staticmethod
        def decode(data):
            return data.data.decode(data.data)

    # There should be no need for the auto-registration in __init_subclass__,
    # and it can add a lot of cruft to the _registries.
    def __init_subclass__(cls):
        pass

    # # TODO: Define schema (should specify that there are two possible forms, and use nptype & _ndim)
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     field_schema.update(type ='array',
    #                         items={'type': 'number'})

    @classmethod
    def validate(cls, value, field=None):
        # NB: We don't need to deal with cases where `value` is serialized data:
        #     Serializer.validate + Data.decode will already have converted that
        #     to an array. What we do need to deal with is coercing values to
        #     the required shape and dtype, and raising an error if that is
        #     not possible.
        if field is None:
            field = SimpleNamespace(name="")

        if json_like(value, cls._registry):
            # Rather than growing the whole type hierarchy so `Array[uint,3]`
            # recognizes `Array` as a type it can deserialize, just so we can
            # reuse super().validate, we just do the deserialization here.
            serialized_cls = cls._registry[value[0]]
            data = value[1]
            if not isinstance(data, serialized_cls.Data):
                # breakpoint()
                # data = serialized_cls.Data.validate(data)
                data = serialized_cls.Data(**data)
                validate_dataclass(data, inplace=True)
            value = serialized_cls.Data.decode(data)
            return cls.validate(value)  # Still validate with `cls`: typically `seralized_cls` is *less* specific, like unspecified `Array°

        elif isinstance(value, cls.Data):
            value = cls.Data.decode(value)
            return cls.validate(value)

        elif isinstance(value, np.ndarray):
            # Don't create a new array unless necessary
            if cls._ndim  is not None and value.ndim != cls._ndim:
                raise TypeError(f"{field.name} expects a variable with "
                                f"{cls._ndim} dimensions.")
            # Issubdtype allows specifying abstract numpy types like 'number', 'floating'
            if cls.nptype is None or np.issubdtype(value.dtype, cls.nptype):
                result = value
            elif np.can_cast(value, cls.nptype):
                nptype = infer_numpy_type_to_cast(cls.nptype, value)
                result = value.astype(nptype)
            else:
                raise TypeError(f"Cannot safely cast '{field.name}' (dtype:  "
                                f"{value.dtype}) to type {cls.nptype}.")
        else:
            result = np.array(value)
            # HACK: Since np.array(…) will accept almost anything, we use
            #       heuristics to try to detect when array construction has
            #       probably failed
            if (isinstance(result, (range, Sequence))
                and any(isinstance(x, Iterable)
                        and not isinstance(x, (np.ndarray, str, bytes))
                        for x in result)):
               # Nested iterables should be unwrapped into an n-d array
               # When this fails (if types or depths are inconsistent), then
               # only the outer level is unwrapped.
               raise TypeError(f"Unable to cast {value} to an array.")
            # Check that array matches expected shape and dtype
            if cls._ndim is not None and result.ndim != cls._ndim:
                raise TypeError(
                    f"The dimensionality of the data (dim: {result.ndim}, "
                    f"shape: {result.shape}) does not correspond to the "
                    f"expected of dimensions ({cls._ndim} for '{field.name}').")
            # Issubdtype allows specifying abstract dtypes like 'number', 'floating'
            if cls.nptype is None or np.issubdtype(result.dtype, cls.nptype):
                pass
            else:
                # Inferring the nptype first avoids warnings if cls.nptype is an abstract type, like np.inexact
                nptype = infer_numpy_type_to_cast(cls.nptype, result)
                if np.can_cast(result, nptype, casting="safe"):
                    result = result.astype(nptype)
                else:
                    raise TypeError(f"Cannot safely cast '{field.name}' (dtype:  "
                                    f"{result.dtype}) to an array of type {cls.nptype}.")
        return result

class _ArrayMeta(type):
    @cache  # Memoization ensures types are created only once
    def __getitem__(self, args):
        if isinstance(args, tuple):
            T = args[0]
            ndim = args[1] if len(args) > 1 else None
            extraargs = args[2:]  # For catching errors only
        else:
            T = args
            ndim = None
            extraargs = []
        if isinstance(T, np.dtype):
            T = T.type
        elif isinstance(T, str):
            T = np.dtype(T).type
        if (not isinstance(T, type) or len(extraargs) > 0
            or not isinstance(ndim, (int, type(None)))):
            # if isinstance(args, (tuple, list)):
            #     argstr = ', '.join((str(a) for a in args))
            # else:
            #     argstr = str(args)
            raise TypeError(
                "`Array` must be specified as either `Array[T]`"
                "or `Array[T, n], where `T` is a type and `n` is an int. "
                f"(received: {args}]).")
        nptype=convert_nptype(T)
        # specifier = str(nptype)
        specifier = nptype.__name__
        if ndim is not None:
            specifier += f",{ndim}"
        # There should be no need for the auto-registration in __init_subclass__,
        # and it can add a lot of cruft to the _registries.
        return type(f'Array[{specifier}]', (_ArrayType,),
                    {'nptype': nptype, '_ndim': ndim})

# NB: We keep Array and _ArrayType separate, so that only Array has the _ArrayMeta metaclass
class Array(_ArrayType, metaclass=_ArrayMeta):
    """
    Use this to specify a NumPy array type annotation; `pydantic` will
    recognize the type and execute appropriate validation/parsing.

    This may become obsolete, or need to be updated, when NumPy officially
    supports type hints (see https://github.com/numpy/numpy-stubs).

    - `Array[T]` specifies an array with dtype `T`. Any expression for which
      `np.dtype(T)` is valid is accepted.
    - `Array[T,n]` specifies an array with dtype `T`, that must have exactly
      `n` dimensions.

    Example
    -------
    >>> from pydantic.dataclasses import dataclass
    >>> from scityping import Array
    >>>
    >>> @dataclass
    >>> class Model:
    >>>     x: Array[np.float64]      # Array of 64-bit floats, any number of dimensions
    >>>     v: Array['float64', 1]    # 1-D array of 64-bit floats
    """
    pass

_ArrayType.register(Array)  # Required because we removed __init_subclass__
_ArrayType.register(np.ndarray)

# ####
# NumPy random generators
#
# Generators are containers around a BitGenerator; it's the state of
# BitGenerator that we need to save. Default BitGenerator is 'PCG64'.
# State is a dict with two required fields: 'bit_generator' and 'state',
# plus 1-3 extra fields depending on the generator.
# 'state' field is itself a dictionary, which may contain arrays of
# type uint32 or uint64

RNGStateDict = Dict[str, Union[int, Array[np.unsignedinteger,1]]]

class RNGenerator(Serializable, np.random.Generator):
    class Data(SerializedData):
        bit_generator: str                       # Philox, PCG64, SFG64, MT19937  
        state        : RNGStateDict              # Philox, PCG64, SFG64, MT19937
        has_uint32   : Optional[int]=None        # Philox, PCG64, SFG64
        uinteger     : Optional[int]=None        # Philox, PCG64, SFG64
        buffer       : Optional[Array[np.uint64,1]]=None  # Philox
        buffer_pos   : Optional[int]=None        # Philox
            # Pydantic will recursively encode state entries, and use Array's
            # reduce when needed
        def encode(rng):
            return rng.bit_generator.state
                # Since this is a dict, RNGenerator.reduce will pass by keyword to `Data`
        def decode(data: 'RNGenerator.Data') -> np.random.Generator:
            bg = getattr(np.random, data.bit_generator)()
            bg.state = asdict(data)
            return np.random.Generator(bg)
    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     field_schema.update(type='array',
    #                         items=[{'type': 'string',
    #                                 'type': 'object'}])

NPGenerator = RNGenerator  # Previous name
Generator = RNGenerator  # To match the name in NumPy

NPGenerator.register(np.random.Generator)

class RandomState(Serializable, np.random.RandomState):
    """Pydantic typing support for the legacy RandomState object.

    .. Note:: Currently tested with the MT19937 generator.
    """
    class Data(SerializedData):
        state: Tuple[str, Array[np.uint64], int, int, float]  # name: str, state: array, int, int, float
        def encode(random_state): return {"state": random_state.get_state()}  # Simply returning tuple would be interpreted as arguments to Data
        def decode(data):
            rs = np.random.RandomState()
            rs.set_state(data.state)
            return rs

    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     field_schema.update(
    #     type='array',
    #     items=[{'type': 'string'},
    #            {'type': 'array',
    #             'items': [{'type': 'array', 'items': {'type': 'integer'}},
    #                       {'type': 'integer'},
    #                       {'type': 'integer'},
    #                       {'type': 'number'}]}
    #           ])

# ####
# NumPy SeedSequence
#
# NOTE: A serialized `SeedSequence` always has a fixed entropy, even when
#    it was created with `entropy=None`. If you need to serialize a `SeedSequence`
#    which returns different values every time it is loaded, you may serialize
#    the arguments and recreate the `SeedSequence` in your code. But also
#    consider that serializing a non-deterministic object is an anti-pattern.

class SeedSequence(Serializable, np.random.SeedSequence):
    class Data(SerializedData):
        entropy           : Union[int,List[int]]
        spawn_key         : Tuple[int,...]
        pool_size         : int
        n_children_spawned: int
        def encode(seedseq) -> SeedSequence.Data:
            return (seedseq.entropy, seedseq.spawn_key,
                    seedseq.pool_size, seedseq.n_children_spawned)

